#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import migrate
import sqlalchemy as sql


ROLE_TABLE = 'role'
IMPLIED_ROLE_TABLE = 'implied_role'


def upgrade(migrate_engine):
    meta = sql.MetaData()
    meta.bind = migrate_engine

    role = sql.Table(ROLE_TABLE, meta, autoload=True)
    implied_role = sql.Table(IMPLIED_ROLE_TABLE, meta, autoload=True)

    fkeys = [
        {'columns': [implied_role.c.prior_role_id],
         'references': [role.c.id]},
        {'columns': [implied_role.c.implied_role_id],
         'references': [role.c.id]},
    ]

    # NOTE(stevemar): We need to divide these into two separate loops otherwise
    # they may clobber each other and only end up with one foreign key.
    for fkey in fkeys:
        migrate.ForeignKeyConstraint(columns=fkey['columns'],
                                     refcolumns=fkey['references'],
                                     name=fkey.get('name')).drop()
    for fkey in fkeys:
        migrate.ForeignKeyConstraint(columns=fkey['columns'],
                                     refcolumns=fkey['references'],
                                     name=fkey.get('name'),
                                     ondelete="CASCADE").create()



# Copyright 2014 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import copy

from testtools import matchers

from keystone.common import json_home
from keystone.tests import unit


class JsonHomeTest(unit.BaseTestCase):
    def test_build_v3_resource_relation(self):
        resource_name = self.getUniqueString()
        relation = json_home.build_v3_resource_relation(resource_name)
        exp_relation = (
            'https://docs.openstack.org/api/openstack-identity/3/rel/%s' %
            resource_name)
        self.assertThat(relation, matchers.Equals(exp_relation))

    def test_build_v3_extension_resource_relation(self):
        extension_name = self.getUniqueString()
        extension_version = self.getUniqueString()
        resource_name = self.getUniqueString()
        relation = json_home.build_v3_extension_resource_relation(
            extension_name, extension_version, resource_name)
        exp_relation = (
            'https://docs.openstack.org/api/openstack-identity/3/ext/%s/%s/rel'
            '/%s' % (extension_name, extension_version, resource_name))
        self.assertThat(relation, matchers.Equals(exp_relation))

    def test_build_v3_parameter_relation(self):
        parameter_name = self.getUniqueString()
        relation = json_home.build_v3_parameter_relation(parameter_name)
        exp_relation = (
            'https://docs.openstack.org/api/openstack-identity/3/param/%s' %
            parameter_name)
        self.assertThat(relation, matchers.Equals(exp_relation))

    def test_build_v3_extension_parameter_relation(self):
        extension_name = self.getUniqueString()
        extension_version = self.getUniqueString()
        parameter_name = self.getUniqueString()
        relation = json_home.build_v3_extension_parameter_relation(
            extension_name, extension_version, parameter_name)
        exp_relation = (
            'https://docs.openstack.org/api/openstack-identity/3/ext/%s/%s/'
            'param/%s' % (extension_name, extension_version, parameter_name))
        self.assertThat(relation, matchers.Equals(exp_relation))

    def test_translate_urls(self):
        href_rel = self.getUniqueString()
        href = self.getUniqueString()
        href_template_rel = self.getUniqueString()
        href_template = self.getUniqueString()
        href_vars = {self.getUniqueString(): self.getUniqueString()}
        original_json_home = {
            'resources': {
                href_rel: {'href': href},
                href_template_rel: {
                    'href-template': href_template,
                    'href-vars': href_vars}
            }
        }

        new_json_home = copy.deepcopy(original_json_home)
        new_prefix = self.getUniqueString()
        json_home.translate_urls(new_json_home, new_prefix)

        exp_json_home = {
            'resources': {
                href_rel: {'href': new_prefix + href},
                href_template_rel: {
                    'href-template': new_prefix + href_template,
                    'href-vars': href_vars}
            }
        }

        self.assertThat(new_json_home, matchers.Equals(exp_json_home))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sqlalchemy as sql

from keystone.common.sql import upgrades


MYSQL_INSERT_TRIGGER = """
CREATE TRIGGER idp_insert_read_only BEFORE INSERT ON identity_provider
FOR EACH ROW
BEGIN
  SIGNAL SQLSTATE '45000'
    SET MESSAGE_TEXT = '%s';
END;
"""

SQLITE_INSERT_TRIGGER = """
CREATE TRIGGER idp_insert_read_only BEFORE INSERT ON identity_provider
BEGIN
  SELECT RAISE (ABORT, '%s');
END;
"""

POSTGRESQL_INSERT_TRIGGER = """
CREATE OR REPLACE FUNCTION keystone_read_only_insert()
  RETURNS trigger AS
$BODY$
BEGIN
  RAISE EXCEPTION '%s';
END
$BODY$ LANGUAGE plpgsql;

CREATE TRIGGER idp_insert_read_only BEFORE INSERT ON identity_provider
FOR EACH ROW
EXECUTE PROCEDURE keystone_read_only_insert();
"""


def upgrade(migrate_engine):
    meta = sql.MetaData()
    meta.bind = migrate_engine

    idp = sql.Table('identity_provider', meta, autoload=True)
    project = sql.Table('project', meta, autoload=True)
    domain_id = sql.Column('domain_id', sql.String(64),
                           sql.ForeignKey(project.c.id), nullable=True)
    idp.create_column(domain_id)

    if upgrades.USE_TRIGGERS:
        # Setting idp to be read-only to prevent old code from creating an idp
        # without a domain_id during an upgrade. This should be okay as it is
        # highly unlikely that an idp would be created during the migration and
        # the impact from preventing creations is minor.
        error_message = ('Identity provider migration in progress. Cannot '
                         'insert new rows into the identity_provider table at '
                         'this time.')
        if migrate_engine.name == 'postgresql':
            idp_insert_trigger = POSTGRESQL_INSERT_TRIGGER % error_message
        elif migrate_engine.name == 'sqlite':
            idp_insert_trigger = SQLITE_INSERT_TRIGGER % error_message
        else:
            idp_insert_trigger = MYSQL_INSERT_TRIGGER % error_message
        migrate_engine.execute(idp_insert_trigger)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Utilities for Federation Extension."""

import ast
import re

import flask
import jsonschema
from oslo_config import cfg
from oslo_log import log
from oslo_utils import timeutils
import six

from keystone.common import provider_api
import keystone.conf
from keystone import exception
from keystone.i18n import _


CONF = keystone.conf.CONF
LOG = log.getLogger(__name__)
PROVIDERS = provider_api.ProviderAPIs


class UserType(object):
    """User mapping type."""

    EPHEMERAL = 'ephemeral'
    LOCAL = 'local'

ROLE_PROPERTIES = {
    "type": "array",
    "items": {
        "type": "object",
        "required": ["name"],
        "properties": {
            "name": {
                "type": "string"
            },
            "additionalProperties": False
        }
    }
}


MAPPING_SCHEMA = {
    "type": "object",
    "required": ['rules'],
    "properties": {
        "rules": {
            "minItems": 1,
            "type": "array",
            "items": {
                "type": "object",
                "required": ['local', 'remote'],
                "additionalProperties": False,
                "properties": {
                    "local": {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "additionalProperties": False,
                            "properties": {
                                "user": {
                                    "type": "object",
                                    "properties": {
                                        "id": {"type": "string"},
                                        "name": {"type": "string"},
                                        "email": {"type": "string"},
                                        "domain": {
                                            "$ref": "#/definitions/domain"
                                        },
                                        "type": {
                                            "type": "string",
                                            "enum": [UserType.EPHEMERAL,
                                                     UserType.LOCAL]
                                        }
                                    },
                                    "additionalProperties": False
                                },
                                "projects": {
                                    "type": "array",
                                    "items": {
                                        "type": "object",
                                        "required": ["name", "roles"],
                                        "additionalProperties": False,
                                        "properties": {
                                            "name": {"type": "string"},
                                            "roles": ROLE_PROPERTIES
                                        }
                                    }
                                },
                                "group": {
                                    "type": "object",
                                    "oneOf": [
                                        {"$ref": "#/definitions/group_by_id"},
                                        {"$ref": "#/definitions/group_by_name"}
                                    ]
                                },
                                "groups": {
                                    "type": "string"
                                },
                                "group_ids": {
                                    "type": "string"
                                },
                                "domain": {"$ref": "#/definitions/domain"},
                            }
                        }
                    },
                    "remote": {
                        "minItems": 1,
                        "type": "array",
                        "items": {
                            "type": "object",
                            "oneOf": [
                                {"$ref": "#/definitions/empty"},
                                {"$ref": "#/definitions/any_one_of"},
                                {"$ref": "#/definitions/not_any_of"},
                                {"$ref": "#/definitions/blacklist"},
                                {"$ref": "#/definitions/whitelist"}
                            ],
                        }
                    }
                }
            }
        }
    },
    "definitions": {
        "empty": {
            "type": "object",
            "required": ['type'],
            "properties": {
                "type": {
                    "type": "string"
                },
            },
            "additionalProperties": False,
        },
        "any_one_of": {
            "type": "object",
            "additionalProperties": False,
            "required": ['type', 'any_one_of'],
            "properties": {
                "type": {
                    "type": "string"
                },
                "any_one_of": {
                    "type": "array"
                },
                "regex": {
                    "type": "boolean"
                }
            }
        },
        "not_any_of": {
            "type": "object",
            "additionalProperties": False,
            "required": ['type', 'not_any_of'],
            "properties": {
                "type": {
                    "type": "string"
                },
                "not_any_of": {
                    "type": "array"
                },
                "regex": {
                    "type": "boolean"
                }
            }
        },
        "blacklist": {
            "type": "object",
            "additionalProperties": False,
            "required": ['type', 'blacklist'],
            "properties": {
                "type": {
                    "type": "string"
                },
                "blacklist": {
                    "type": "array"
                }
            }
        },
        "whitelist": {
            "type": "object",
            "additionalProperties": False,
            "required": ['type', 'whitelist'],
            "properties": {
                "type": {
                    "type": "string"
                },
                "whitelist": {
                    "type": "array"
                }
            }
        },
        "domain": {
            "type": "object",
            "properties": {
                "id": {"type": "string"},
                "name": {"type": "string"}
            },
            "additionalProperties": False
        },
        "group_by_id": {
            "type": "object",
            "properties": {
                "id": {"type": "string"}
            },
            "additionalProperties": False,
            "required": ["id"]
        },
        "group_by_name": {
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "domain": {"$ref": "#/definitions/domain"}
            },
            "additionalProperties": False,
            "required": ["name", "domain"]
        }
    }
}


class DirectMaps(object):
    """An abstraction around the remote matches.

    Each match is treated internally as a list.
    """

    def __init__(self):
        self._matches = []

    def __str__(self):
        """return the direct map array as a string."""
        return '%s' % self._matches

    def add(self, values):
        """Add a matched value to the list of matches.

        :param list value: the match to save

        """
        self._matches.append(values)

    def __getitem__(self, idx):
        """Used by Python when executing ``''.format(*DirectMaps())``."""
        value = self._matches[idx]
        if isinstance(value, list) and len(value) == 1:
            return value[0]
        else:
            return value


def validate_mapping_structure(ref):
    v = jsonschema.Draft4Validator(MAPPING_SCHEMA)

    messages = ''
    for error in sorted(v.iter_errors(ref), key=str):
        messages = messages + error.message + "\n"

    if messages:
        raise exception.ValidationError(messages)


def validate_expiration(token):
    token_expiration_datetime = timeutils.normalize_time(
        timeutils.parse_isotime(token.expires_at)
    )
    if timeutils.utcnow() > token_expiration_datetime:
        raise exception.Unauthorized(_('Federation token is expired'))


def get_remote_id_parameter(idp, protocol):
    # NOTE(marco-fargetta): Since we support any protocol ID, we attempt to
    # retrieve the remote_id_attribute of the protocol ID. It will look up first
    # if the remote_id_attribute exists.
    protocol_ref = PROVIDERS.federation_api.get_protocol(idp['id'], protocol)
    remote_id_parameter = protocol_ref.get('remote_id_attribute')
    if remote_id_parameter:
        return remote_id_parameter
    else:
        # If it's not registered in the config, then register the option and try again.
        # This allows the user to register protocols other than oidc and saml2.
        try:
            remote_id_parameter = CONF[protocol]['remote_id_attribute']
        except AttributeError:
            # TODO(dolph): Move configuration registration to keystone.conf
            CONF.register_opt(cfg.StrOpt('remote_id_attribute'),
                              group=protocol)
            try:
                remote_id_parameter = CONF[protocol]['remote_id_attribute']
            except AttributeError:  # nosec
                # No remote ID attr, will be logged and use the default instead.
                pass
    if not remote_id_parameter:
        LOG.debug('Cannot find "remote_id_attribute" in configuration '
                  'group %s. Trying default location in '
                  'group federation.', protocol)
        remote_id_parameter = CONF.federation.remote_id_attribute

    return remote_id_parameter


def validate_idp(idp, protocol, assertion):
    """The IdP providing the assertion should be registered for the mapping."""
    remote_id_parameter = get_remote_id_parameter(idp, protocol)
    if not remote_id_parameter or not idp['remote_ids']:
        LOG.debug('Impossible to identify the IdP %s ', idp['id'])
        # If nothing is defined, the administrator may want to
        # allow the mapping of every IdP
        return
    try:
        idp_remote_identifier = assertion[remote_id_parameter]
    except KeyError:
        msg = _('Could not find Identity Provider identifier in '
                'environment')
        raise exception.ValidationError(msg)
    if idp_remote_identifier not in idp['remote_ids']:
        msg = _('Incoming identity provider identifier not included '
                'among the accepted identifiers.')
        raise exception.Forbidden(msg)


def validate_mapped_group_ids(group_ids, mapping_id, identity_api):
    """Iterate over group ids and make sure they are present in the backend.

    This call is not transactional.
    :param group_ids: IDs of the groups to be checked
    :type group_ids: list of str

    :param mapping_id: id of the mapping used for this operation
    :type mapping_id: str

    :param identity_api: Identity Manager object used for communication with
                         backend
    :type identity_api: identity.Manager

    :raises keystone.exception.MappedGroupNotFound: If the group returned by
        mapping was not found in the backend.

    """
    for group_id in group_ids:
        try:
            identity_api.get_group(group_id)
        except exception.GroupNotFound:
            raise exception.MappedGroupNotFound(
                group_id=group_id, mapping_id=mapping_id)


# TODO(marek-denis): Optimize this function, so the number of calls to the
# backend are minimized.
def transform_to_group_ids(group_names, mapping_id,
                           identity_api, resource_api):
    """Transform groups identified by name/domain to their ids.

    Function accepts list of groups identified by a name and domain giving
    a list of group ids in return. A message is logged if the group doesn't
    exist in the backend.

    Example of group_names parameter::

        [
            {
                "name": "group_name",
                "domain": {
                    "id": "domain_id"
                },
            },
            {
                "name": "group_name_2",
                "domain": {
                    "name": "domain_name"
                }
            }
        ]

    :param group_names: list of group identified by name and its domain.
    :type group_names: list

    :param mapping_id: id of the mapping used for mapping assertion into
        local credentials
    :type mapping_id: str

    :param identity_api: identity_api object
    :param resource_api: resource manager object

    :returns: generator object with group ids

    """
    def resolve_domain(domain):
        """Return domain id.

        Input is a dictionary with a domain identified either by a ``id`` or a
        ``name``. In the latter case system will attempt to fetch domain object
        from the backend.

        :returns: domain's id
        :rtype: str

        """
        domain_id = (domain.get('id') or
                     resource_api.get_domain_by_name(
                     domain.get('name')).get('id'))
        return domain_id

    for group in group_names:
        try:
            group_dict = identity_api.get_group_by_name(
                group['name'], resolve_domain(group['domain']))
            yield group_dict['id']
        except exception.GroupNotFound:
            LOG.debug('Group %s has no entry in the backend',
                      group['name'])


def get_assertion_params_from_env():
    LOG.debug('Environment variables: %s', flask.request.environ)
    prefix = CONF.federation.assertion_prefix
    for k, v in list(flask.request.environ.items()):
        if not k.startswith(prefix):
            continue
        # These bytes may be decodable as ISO-8859-1 according to Section
        # 3.2.4 of RFC 7230. Let's assume that our web server plugins are
        # correctly encoding the data.
        if not isinstance(v, six.text_type) and getattr(v, 'decode', False):
            v = v.decode('ISO-8859-1')
        yield (k, v)


class RuleProcessor(object):
    """A class to process assertions and mapping rules."""

    class _EvalType(object):
        """Mapping rule evaluation types."""

        ANY_ONE_OF = 'any_one_of'
        NOT_ANY_OF = 'not_any_of'
        BLACKLIST = 'blacklist'
        WHITELIST = 'whitelist'

    def __init__(self, mapping_id, rules):
        """Initialize RuleProcessor.

        Example rules can be found at:
        :class:`keystone.tests.mapping_fixtures`

        :param mapping_id: id for the mapping
        :type mapping_id: string
        :param rules: rules from a mapping
        :type rules: dict

        """
        self.mapping_id = mapping_id
        self.rules = rules

    def process(self, assertion_data):
        """Transform assertion to a dictionary.

        The dictionary contains mapping of user name and group ids
        based on mapping rules.

        This function will iterate through the mapping rules to find
        assertions that are valid.

        :param assertion_data: an assertion containing values from an IdP
        :type assertion_data: dict

        Example assertion_data::

            {
                'Email': 'testacct@example.com',
                'UserName': 'testacct',
                'FirstName': 'Test',
                'LastName': 'Account',
                'orgPersonType': 'Tester'
            }

        :returns: dictionary with user and group_ids

        The expected return structure is::

            {
                'name': 'foobar',
                'group_ids': ['abc123', 'def456'],
                'group_names': [
                    {
                        'name': 'group_name_1',
                        'domain': {
                            'name': 'domain1'
                        }
                    },
                    {
                        'name': 'group_name_1_1',
                        'domain': {
                            'name': 'domain1'
                        }
                    },
                    {
                        'name': 'group_name_2',
                        'domain': {
                            'id': 'xyz132'
                        }
                    }
                ]
            }

        """
        # Assertions will come in as string key-value pairs, and will use a
        # semi-colon to indicate multiple values, i.e. groups.
        # This will create a new dictionary where the values are arrays, and
        # any multiple values are stored in the arrays.
        LOG.debug('assertion data: %s', assertion_data)
        assertion = {n: v.split(';') for n, v in assertion_data.items()
                     if isinstance(v, six.string_types)}
        LOG.debug('assertion: %s', assertion)
        identity_values = []

        LOG.debug('rules: %s', self.rules)
        for rule in self.rules:
            direct_maps = self._verify_all_requirements(rule['remote'],
                                                        assertion)

            # If the compare comes back as None, then the rule did not apply
            # to the assertion data, go on to the next rule
            if direct_maps is None:
                continue

            # If there are no direct mappings, then add the local mapping
            # directly to the array of saved values. However, if there is
            # a direct mapping, then perform variable replacement.
            if not direct_maps:
                identity_values += rule['local']
            else:
                for local in rule['local']:
                    new_local = self._update_local_mapping(local, direct_maps)
                    identity_values.append(new_local)

        LOG.debug('identity_values: %s', identity_values)
        mapped_properties = self._transform(identity_values)
        LOG.debug('mapped_properties: %s', mapped_properties)
        return mapped_properties

    def _transform(self, identity_values):
        """Transform local mappings, to an easier to understand format.

        Transform the incoming array to generate the return value for
        the process function. Generating content for Keystone tokens will
        be easier if some pre-processing is done at this level.

        :param identity_values: local mapping from valid evaluations
        :type identity_values: array of dict

        Example identity_values::

            [
                {
                    'group': {'id': '0cd5e9'},
                    'user': {
                        'email': 'bob@example.com'
                    },
                },
                {
                    'groups': ['member', 'admin', tester'],
                    'domain': {
                        'name': 'default_domain'
                    }
                },
                {
                    'group_ids': ['abc123', 'def456', '0cd5e9']
                }
            ]

        :returns: dictionary with user name, group_ids and group_names.
        :rtype: dict

        """
        def extract_groups(groups_by_domain):
            for groups in list(groups_by_domain.values()):
                for group in list({g['name']: g for g in groups}.values()):
                    yield group

        def normalize_user(user):
            """Parse and validate user mapping."""
            user_type = user.get('type')

            if user_type and user_type not in (UserType.EPHEMERAL,
                                               UserType.LOCAL):
                msg = _("User type %s not supported") % user_type
                raise exception.ValidationError(msg)

            if user_type is None:
                user['type'] = UserType.EPHEMERAL

        # initialize the group_ids as a set to eliminate duplicates
        user = {}
        group_ids = set()
        group_names = list()
        groups_by_domain = dict()
        projects = []

        # if mapping yield no valid identity values, we should bail right away
        # instead of continuing on with a normalized bogus user
        if not identity_values:
            msg = ("Could not map any federated user properties to identity "
                   "values. Check debug logs or the mapping used for "
                   "additional details.")
            tr_msg = _("Could not map any federated user properties to "
                       "identity values. Check debug logs or the mapping "
                       "used for additional details.")
            LOG.warning(msg)
            raise exception.ValidationError(tr_msg)

        for identity_value in identity_values:
            if 'user' in identity_value:
                # if a mapping outputs more than one user name, log it
                if user:
                    LOG.warning('Ignoring user name')
                else:
                    user = identity_value.get('user')
            if 'group' in identity_value:
                group = identity_value['group']
                if 'id' in group:
                    group_ids.add(group['id'])
                elif 'name' in group:
                    domain = (group['domain'].get('name') or
                              group['domain'].get('id'))
                    groups_by_domain.setdefault(domain, list()).append(group)
                group_names.extend(extract_groups(groups_by_domain))
            if 'groups' in identity_value:
                if 'domain' not in identity_value:
                    msg = _("Invalid rule: %(identity_value)s. Both 'groups' "
                            "and 'domain' keywords must be specified.")
                    msg = msg % {'identity_value': identity_value}
                    raise exception.ValidationError(msg)
                # In this case, identity_value['groups'] is a string
                # representation of a list, and we want a real list.  This is
                # due to the way we do direct mapping substitutions today (see
                # function _update_local_mapping() )
                try:
                    group_names_list = ast.literal_eval(
                        identity_value['groups'])
                except (ValueError, SyntaxError):
                    group_names_list = [identity_value['groups']]
                domain = identity_value['domain']
                group_dicts = [{'name': name, 'domain': domain} for name in
                               group_names_list]

                group_names.extend(group_dicts)
            if 'group_ids' in identity_value:
                # If identity_values['group_ids'] is a string representation
                # of a list, parse it to a real list. Also, if the provided
                # group_ids parameter contains only one element, it will be
                # parsed as a simple string, and not a list or the
                # representation of a list.
                try:
                    group_ids.update(
                        ast.literal_eval(identity_value['group_ids']))
                except (ValueError, SyntaxError):
                    group_ids.update([identity_value['group_ids']])
            if 'projects' in identity_value:
                projects = identity_value['projects']

        normalize_user(user)

        return {'user': user,
                'group_ids': list(group_ids),
                'group_names': group_names,
                'projects': projects}

    def _update_local_mapping(self, local, direct_maps):
        """Replace any {0}, {1} ... values with data from the assertion.

        :param local: local mapping reference that needs to be updated
        :type local: dict
        :param direct_maps: identity values used to update local
        :type direct_maps: keystone.federation.utils.DirectMaps

        Example local::

            {'user': {'name': '{0} {1}', 'email': '{2}'}}

        Example direct_maps::

            [['Bob'], ['Thompson'], ['bob@example.com']]

        :returns: new local mapping reference with replaced values.

        The expected return structure is::

            {'user': {'name': 'Bob Thompson', 'email': 'bob@example.org'}}

        :raises keystone.exception.DirectMappingError: when referring to a
            remote match from a local section of a rule

        """
        LOG.debug('direct_maps: %s', direct_maps)
        LOG.debug('local: %s', local)
        new = {}
        for k, v in local.items():
            if isinstance(v, dict):
                new_value = self._update_local_mapping(v, direct_maps)
            elif isinstance(v, list):
                new_value = [self._update_local_mapping(item, direct_maps)
                             for item in v]
            else:
                try:
                    new_value = v.format(*direct_maps)
                except IndexError:
                    raise exception.DirectMappingError(
                        mapping_id=self.mapping_id)

            new[k] = new_value
        return new

    def _verify_all_requirements(self, requirements, assertion):
        """Compare remote requirements of a rule against the assertion.

        If a value of ``None`` is returned, the rule with this assertion
        doesn't apply.
        If an array of zero length is returned, then there are no direct
        mappings to be performed, but the rule is valid.
        Otherwise, then it will first attempt to filter the values according
        to blacklist or whitelist rules and finally return the values in
        order, to be directly mapped.

        :param requirements: list of remote requirements from rules
        :type requirements: list

        Example requirements::

            [
                {
                    "type": "UserName"
                },
                {
                    "type": "orgPersonType",
                    "any_one_of": [
                        "Customer"
                    ]
                },
                {
                    "type": "ADFS_GROUPS",
                    "whitelist": [
                        "g1", "g2", "g3", "g4"
                    ]
                }
            ]

        :param assertion: dict of attributes from an IdP
        :type assertion: dict

        Example assertion::

            {
                'UserName': ['testacct'],
                'LastName': ['Account'],
                'orgPersonType': ['Tester'],
                'Email': ['testacct@example.com'],
                'FirstName': ['Test'],
                'ADFS_GROUPS': ['g1', 'g2']
            }

        :returns: identity values used to update local
        :rtype: keystone.federation.utils.DirectMaps or None

        """
        direct_maps = DirectMaps()

        for requirement in requirements:
            requirement_type = requirement['type']
            direct_map_values = assertion.get(requirement_type)
            regex = requirement.get('regex', False)

            if not direct_map_values:
                return None

            any_one_values = requirement.get(self._EvalType.ANY_ONE_OF)
            if any_one_values is not None:
                if self._evaluate_requirement(any_one_values,
                                              direct_map_values,
                                              self._EvalType.ANY_ONE_OF,
                                              regex):
                    continue
                else:
                    return None

            not_any_values = requirement.get(self._EvalType.NOT_ANY_OF)
            if not_any_values is not None:
                if self._evaluate_requirement(not_any_values,
                                              direct_map_values,
                                              self._EvalType.NOT_ANY_OF,
                                              regex):
                    continue
                else:
                    return None

            # If 'any_one_of' or 'not_any_of' are not found, then values are
            # within 'type'. Attempt to find that 'type' within the assertion,
            # and filter these values if 'whitelist' or 'blacklist' is set.
            blacklisted_values = requirement.get(self._EvalType.BLACKLIST)
            whitelisted_values = requirement.get(self._EvalType.WHITELIST)

            # If a blacklist or whitelist is used, we want to map to the
            # whole list instead of just its values separately.
            if blacklisted_values is not None:
                direct_map_values = [v for v in direct_map_values
                                     if v not in blacklisted_values]
            elif whitelisted_values is not None:
                direct_map_values = [v for v in direct_map_values
                                     if v in whitelisted_values]

            direct_maps.add(direct_map_values)

            LOG.debug('updating a direct mapping: %s', direct_map_values)

        return direct_maps

    def _evaluate_values_by_regex(self, values, assertion_values):
        for value in values:
            for assertion_value in assertion_values:
                if re.search(value, assertion_value):
                    return True
        return False

    def _evaluate_requirement(self, values, assertion_values,
                              eval_type, regex):
        """Evaluate the incoming requirement and assertion.

        If the requirement type does not exist in the assertion data, then
        return False. If regex is specified, then compare the values and
        assertion values. Otherwise, grab the intersection of the values
        and use that to compare against the evaluation type.

        :param values: list of allowed values, defined in the requirement
        :type values: list
        :param assertion_values: The values from the assertion to evaluate
        :type assertion_values: list/string
        :param eval_type: determine how to evaluate requirements
        :type eval_type: string
        :param regex: perform evaluation with regex
        :type regex: boolean

        :returns: boolean, whether requirement is valid or not.

        """
        if regex:
            any_match = self._evaluate_values_by_regex(values,
                                                       assertion_values)
        else:
            any_match = bool(set(values).intersection(set(assertion_values)))
        if any_match and eval_type == self._EvalType.ANY_ONE_OF:
            return True
        if not any_match and eval_type == self._EvalType.NOT_ANY_OF:
            return True

        return False


def assert_enabled_identity_provider(federation_api, idp_id):
    identity_provider = federation_api.get_idp(idp_id)
    if identity_provider.get('enabled') is not True:
        msg = 'Identity Provider %(idp)s is disabled' % {
            'idp': idp_id}
        tr_msg = _('Identity Provider %(idp)s is disabled') % {
            'idp': idp_id}
        LOG.debug(msg)
        raise exception.Forbidden(tr_msg)


def assert_enabled_service_provider_object(service_provider):
    if service_provider.get('enabled') is not True:
        sp_id = service_provider['id']
        msg = 'Service Provider %(sp)s is disabled' % {'sp': sp_id}
        tr_msg = _('Service Provider %(sp)s is disabled') % {'sp': sp_id}
        LOG.debug(msg)
        raise exception.Forbidden(tr_msg)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import migrate
from oslo_log import log
import sqlalchemy as sql

from keystone.assignment.backends import sql as assignment_sql
from keystone.common import sql as ks_sql
from keystone.identity.mapping_backends import mapping as mapping_backend


LOG = log.getLogger(__name__)


def upgrade(migrate_engine):
    meta = sql.MetaData()
    meta.bind = migrate_engine

    if migrate_engine.name == 'mysql':
        # In Folsom we explicitly converted migrate_version to UTF8.
        migrate_engine.execute(
            'ALTER TABLE migrate_version CONVERT TO CHARACTER SET utf8')
        # Set default DB charset to UTF8.
        migrate_engine.execute(
            'ALTER DATABASE %s DEFAULT CHARACTER SET utf8' %
            migrate_engine.url.database)

    credential = sql.Table(
        'credential', meta,
        sql.Column('id', sql.String(length=64), primary_key=True),
        sql.Column('user_id', sql.String(length=64), nullable=False),
        sql.Column('project_id', sql.String(length=64)),
        sql.Column('blob', ks_sql.JsonBlob, nullable=False),
        sql.Column('type', sql.String(length=255), nullable=False),
        sql.Column('extra', ks_sql.JsonBlob.impl),
        mysql_engine='InnoDB',
        mysql_charset='utf8')

    domain = sql.Table(
        'domain', meta,
        sql.Column('id', sql.String(length=64), primary_key=True),
        sql.Column('name', sql.String(length=64), nullable=False),
        sql.Column('enabled', sql.Boolean, default=True, nullable=False),
        sql.Column('extra', ks_sql.JsonBlob.impl),
        mysql_engine='InnoDB',
        mysql_charset='utf8')

    endpoint = sql.Table(
        'endpoint', meta,
        sql.Column('id', sql.String(length=64), primary_key=True),
        sql.Column('legacy_endpoint_id', sql.String(length=64)),
        sql.Column('interface', sql.String(length=8), nullable=False),
        sql.Column('service_id', sql.String(length=64), nullable=False),
        sql.Column('url', sql.Text, nullable=False),
        sql.Column('extra', ks_sql.JsonBlob.impl),
        sql.Column('enabled', sql.Boolean, nullable=False, default=True,
                   server_default='1'),
        sql.Column('region_id', sql.String(length=255), nullable=True),
        mysql_engine='InnoDB',
        mysql_charset='utf8')

    group = sql.Table(
        'group', meta,
        sql.Column('id', sql.String(length=64), primary_key=True),
        sql.Column('domain_id', sql.String(length=64), nullable=False),
        sql.Column('name', sql.String(length=64), nullable=False),
        sql.Column('description', sql.Text),
        sql.Column('extra', ks_sql.JsonBlob.impl),
        mysql_engine='InnoDB',
        mysql_charset='utf8')

    policy = sql.Table(
        'policy', meta,
        sql.Column('id', sql.String(length=64), primary_key=True),
        sql.Column('type', sql.String(length=255), nullable=False),
        sql.Column('blob', ks_sql.JsonBlob, nullable=False),
        sql.Column('extra', ks_sql.JsonBlob.impl),
        mysql_engine='InnoDB',
        mysql_charset='utf8')

    project = sql.Table(
        'project', meta,
        sql.Column('id', sql.String(length=64), primary_key=True),
        sql.Column('name', sql.String(length=64), nullable=False),
        sql.Column('extra', ks_sql.JsonBlob.impl),
        sql.Column('description', sql.Text),
        sql.Column('enabled', sql.Boolean),
        sql.Column('domain_id', sql.String(length=64), nullable=False),
        sql.Column('parent_id', sql.String(64), nullable=True),
        mysql_engine='InnoDB',
        mysql_charset='utf8')

    role = sql.Table(
        'role', meta,
        sql.Column('id', sql.String(length=64), primary_key=True),
        sql.Column('name', sql.String(length=255), nullable=False),
        sql.Column('extra', ks_sql.JsonBlob.impl),
        mysql_engine='InnoDB',
        mysql_charset='utf8')

    service = sql.Table(
        'service', meta,
        sql.Column('id', sql.String(length=64), primary_key=True),
        sql.Column('type', sql.String(length=255)),
        sql.Column('enabled', sql.Boolean, nullable=False, default=True,
                   server_default='1'),
        sql.Column('extra', ks_sql.JsonBlob.impl),
        mysql_engine='InnoDB',
        mysql_charset='utf8')

    token = sql.Table(
        'token', meta,
        sql.Column('id', sql.String(length=64), primary_key=True),
        sql.Column('expires', sql.DateTime, default=None),
        sql.Column('extra', ks_sql.JsonBlob.impl),
        sql.Column('valid', sql.Boolean, default=True, nullable=False),
        sql.Column('trust_id', sql.String(length=64)),
        sql.Column('user_id', sql.String(length=64)),
        mysql_engine='InnoDB',
        mysql_charset='utf8')

    trust = sql.Table(
        'trust', meta,
        sql.Column('id', sql.String(length=64), primary_key=True),
        sql.Column('trustor_user_id', sql.String(length=64), nullable=False),
        sql.Column('trustee_user_id', sql.String(length=64), nullable=False),
        sql.Column('project_id', sql.String(length=64)),
        sql.Column('impersonation', sql.Boolean, nullable=False),
        sql.Column('deleted_at', sql.DateTime),
        sql.Column('expires_at', sql.DateTime),
        sql.Column('remaining_uses', sql.Integer, nullable=True),
        sql.Column('extra', ks_sql.JsonBlob.impl),
        mysql_engine='InnoDB',
        mysql_charset='utf8')

    trust_role = sql.Table(
        'trust_role', meta,
        sql.Column('trust_id', sql.String(length=64), primary_key=True,
                   nullable=False),
        sql.Column('role_id', sql.String(length=64), primary_key=True,
                   nullable=False),
        mysql_engine='InnoDB',
        mysql_charset='utf8')

    user = sql.Table(
        'user', meta,
        sql.Column('id', sql.String(length=64), primary_key=True),
        sql.Column('name', sql.String(length=255), nullable=False),
        sql.Column('extra', ks_sql.JsonBlob.impl),
        sql.Column('password', sql.String(length=128)),
        sql.Column('enabled', sql.Boolean),
        sql.Column('domain_id', sql.String(length=64), nullable=False),
        sql.Column('default_project_id', sql.String(length=64)),
        mysql_engine='InnoDB',
        mysql_charset='utf8')

    user_group_membership = sql.Table(
        'user_group_membership', meta,
        sql.Column('user_id', sql.String(length=64), primary_key=True),
        sql.Column('group_id', sql.String(length=64), primary_key=True),
        mysql_engine='InnoDB',
        mysql_charset='utf8')

    region = sql.Table(
        'region',
        meta,
        sql.Column('id', sql.String(255), primary_key=True),
        sql.Column('description', sql.String(255), nullable=False),
        sql.Column('parent_region_id', sql.String(255), nullable=True),
        sql.Column('extra', sql.Text()),
        mysql_engine='InnoDB',
        mysql_charset='utf8')

    assignment = sql.Table(
        'assignment',
        meta,
        sql.Column('type', sql.Enum(
            assignment_sql.AssignmentType.USER_PROJECT,
            assignment_sql.AssignmentType.GROUP_PROJECT,
            assignment_sql.AssignmentType.USER_DOMAIN,
            assignment_sql.AssignmentType.GROUP_DOMAIN,
            name='type'),
            nullable=False),
        sql.Column('actor_id', sql.String(64), nullable=False),
        sql.Column('target_id', sql.String(64), nullable=False),
        sql.Column('role_id', sql.String(64), nullable=False),
        sql.Column('inherited', sql.Boolean, default=False, nullable=False),
        sql.PrimaryKeyConstraint('type', 'actor_id', 'target_id', 'role_id'),
        mysql_engine='InnoDB',
        mysql_charset='utf8')

    mapping = sql.Table(
        'id_mapping',
        meta,
        sql.Column('public_id', sql.String(64), primary_key=True),
        sql.Column('domain_id', sql.String(64), nullable=False),
        sql.Column('local_id', sql.String(64), nullable=False),
        sql.Column('entity_type', sql.Enum(
            mapping_backend.EntityType.USER,
            mapping_backend.EntityType.GROUP,
            name='entity_type'),
            nullable=False),
        mysql_engine='InnoDB',
        mysql_charset='utf8')

    domain_config_whitelist = sql.Table(
        'whitelisted_config',
        meta,
        sql.Column('domain_id', sql.String(64), primary_key=True),
        sql.Column('group', sql.String(255), primary_key=True),
        sql.Column('option', sql.String(255), primary_key=True),
        sql.Column('value', ks_sql.JsonBlob.impl, nullable=False),
        mysql_engine='InnoDB',
        mysql_charset='utf8')

    domain_config_sensitive = sql.Table(
        'sensitive_config',
        meta,
        sql.Column('domain_id', sql.String(64), primary_key=True),
        sql.Column('group', sql.String(255), primary_key=True),
        sql.Column('option', sql.String(255), primary_key=True),
        sql.Column('value', ks_sql.JsonBlob.impl, nullable=False),
        mysql_engine='InnoDB',
        mysql_charset='utf8')

    # create all tables
    tables = [credential, domain, endpoint, group, policy, project, role,
              service, token, trust, trust_role, user, user_group_membership,
              region, assignment, mapping, domain_config_whitelist,
              domain_config_sensitive]

    for table in tables:
        try:
            table.create()
        except Exception:
            LOG.exception('Exception while creating table: %r', table)
            raise

    # Unique Constraints
    migrate.UniqueConstraint(user.c.domain_id,
                             user.c.name,
                             name='ixu_user_name_domain_id').create()
    migrate.UniqueConstraint(group.c.domain_id,
                             group.c.name,
                             name='ixu_group_name_domain_id').create()
    migrate.UniqueConstraint(role.c.name,
                             name='ixu_role_name').create()
    migrate.UniqueConstraint(project.c.domain_id,
                             project.c.name,
                             name='ixu_project_name_domain_id').create()
    migrate.UniqueConstraint(domain.c.name,
                             name='ixu_domain_name').create()
    migrate.UniqueConstraint(mapping.c.domain_id,
                             mapping.c.local_id,
                             mapping.c.entity_type,
                             name='domain_id').create()

    # Indexes
    sql.Index('ix_token_expires', token.c.expires).create()
    sql.Index('ix_token_expires_valid', token.c.expires,
              token.c.valid).create()
    sql.Index('ix_actor_id', assignment.c.actor_id).create()
    sql.Index('ix_token_user_id', token.c.user_id).create()
    sql.Index('ix_token_trust_id', token.c.trust_id).create()
    # NOTE(stevemar): The two indexes below were named 'service_id' and
    # 'group_id' in 050_fk_consistent_indexes.py, and need to be preserved
    sql.Index('service_id', endpoint.c.service_id).create()
    sql.Index('group_id', user_group_membership.c.group_id).create()

    fkeys = [
        {'columns': [endpoint.c.service_id],
         'references': [service.c.id]},

        {'columns': [user_group_membership.c.group_id],
         'references': [group.c.id],
         'name': 'fk_user_group_membership_group_id'},

        {'columns': [user_group_membership.c.user_id],
         'references':[user.c.id],
         'name': 'fk_user_group_membership_user_id'},

        {'columns': [project.c.domain_id],
         'references': [domain.c.id],
         'name': 'fk_project_domain_id'},

        {'columns': [endpoint.c.region_id],
         'references': [region.c.id],
         'name': 'fk_endpoint_region_id'},

        {'columns': [project.c.parent_id],
         'references': [project.c.id],
         'name': 'project_parent_id_fkey'},
    ]

    if migrate_engine.name == 'sqlite':
        # NOTE(stevemar): We need to keep this FK constraint due to 073, but
        # only for sqlite, once we collapse 073 we can remove this constraint
        fkeys.append(
            {'columns': [assignment.c.role_id],
             'references': [role.c.id],
             'name': 'fk_assignment_role_id'})

    for fkey in fkeys:
        migrate.ForeignKeyConstraint(columns=fkey['columns'],
                                     refcolumns=fkey['references'],
                                     name=fkey.get('name')).create()



# Copyright 2019 SUSE Linux GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sqlalchemy as sql


def upgrade(migrate_engine):
    meta = sql.MetaData()
    meta.bind = migrate_engine

    application_credential = sql.Table('application_credential', meta, autoload=True)
    access_rule = sql.Table(
        'access_rule', meta,
        sql.Column('id', sql.Integer, primary_key=True, nullable=False),
        sql.Column('service', sql.String(64)),
        sql.Column('path', sql.String(128)),
        sql.Column('method', sql.String(16)),
        mysql_engine='InnoDB', mysql_charset='utf8'
    )
    app_cred_access_rule = sql.Table(
        'application_credential_access_rule', meta,
        sql.Column('application_credential_id', sql.Integer,
                   sql.ForeignKey(application_credential.c.internal_id,
                                  ondelete='CASCADE'),
                   primary_key=True, nullable=False),
        sql.Column('access_rule_id', sql.Integer,
                   sql.ForeignKey(access_rule.c.id,
                                  ondelete='CASCADE'),
                   primary_key=True, nullable=False),
        mysql_engine='InnoDB', mysql_charset='utf8'
    )
    access_rule.create(migrate_engine, checkfirst=True)
    app_cred_access_rule.create(migrate_engine, checkfirst=True)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystone.common import validation
from keystone.common.validation import parameter_types

_role_properties = {
    'type': 'array',
    'items': {
        'type': 'object',
        'properties': {
            'id': parameter_types.id_string,
            'name': parameter_types.name
        },
        'minProperties': 1,
        'maxProperties': 1,
        'additionalProperties': False
    }
}

_trust_properties = {
    # NOTE(lbragstad): These are set as external_id_string because they have
    # the ability to be read as LDAP user identifiers, which could be something
    # other than uuid.
    'trustor_user_id': parameter_types.external_id_string,
    'trustee_user_id': parameter_types.external_id_string,
    'impersonation': parameter_types.boolean,
    'project_id': validation.nullable(parameter_types.id_string),
    'remaining_uses': {
        'type': ['integer', 'null'],
        'minimum': 1
    },
    'expires_at': {
        'type': ['null', 'string']
    },
    'allow_redelegation': {
        'type': ['boolean', 'null']
    },
    'redelegation_count': {
        'type': ['integer', 'null'],
        'minimum': 0
    },
    'roles': _role_properties
}

trust_create = {
    'type': 'object',
    'properties': _trust_properties,
    'required': ['trustor_user_id', 'trustee_user_id', 'impersonation'],
    'additionalProperties': True
}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from keystone.conf import utils


driver = cfg.StrOpt(
    'driver',
    default='sql',
    help=utils.fmt("""
Entry point for the application credential backend driver in the
`keystone.application_credential` namespace.  Keystone only provides a `sql`
driver, so there is no reason to change this unless you are providing a custom
entry point.
"""))

caching = cfg.BoolOpt(
    'caching',
    default=True,
    help=utils.fmt("""
Toggle for application credential caching. This has no effect unless global
caching is enabled.
"""))

cache_time = cfg.IntOpt(
    'cache_time',
    help=utils.fmt("""
Time to cache application credential data in seconds. This has no effect
unless global caching is enabled.
"""))

user_limit = cfg.IntOpt(
    'user_limit',
    default=-1,
    help=utils.fmt("""
Maximum number of application credentials a user is permitted to create. A
value of -1 means unlimited. If a limit is not set, users are permitted to
create application credentials at will, which could lead to bloat in the
keystone database or open keystone to a DoS attack.
"""))


GROUP_NAME = __name__.split('.')[-1]
ALL_OPTS = [
    driver,
    caching,
    cache_time,
    user_limit,
]


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=GROUP_NAME)


def list_opts():
    return {GROUP_NAME: ALL_OPTS}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from keystone.conf import utils


template_file = cfg.StrOpt(
    'template_file',
    default='default_catalog.templates',
    help=utils.fmt("""
Absolute path to the file used for the templated catalog backend. This option
is only used if the `[catalog] driver` is set to `templated`.
"""))

driver = cfg.StrOpt(
    'driver',
    default='sql',
    help=utils.fmt("""
Entry point for the catalog driver in the `keystone.catalog` namespace.
Keystone provides a `sql` option (which supports basic CRUD operations through
SQL), a `templated` option (which loads the catalog from a templated catalog
file on disk), and a `endpoint_filter.sql` option (which supports arbitrary
service catalogs per project).
"""))

caching = cfg.BoolOpt(
    'caching',
    default=True,
    help=utils.fmt("""
Toggle for catalog caching. This has no effect unless global caching is
enabled. In a typical deployment, there is no reason to disable this.
"""))

cache_time = cfg.IntOpt(
    'cache_time',
    help=utils.fmt("""
Time to cache catalog data (in seconds). This has no effect unless global and
catalog caching are both enabled. Catalog data (services, endpoints, etc.)
typically does not change frequently, and so a longer duration than the global
default may be desirable.
"""))

list_limit = cfg.IntOpt(
    'list_limit',
    help=utils.fmt("""
Maximum number of entities that will be returned in a catalog collection. There
is typically no reason to set this, as it would be unusual for a deployment to
have enough services or endpoints to exceed a reasonable limit.
"""))


GROUP_NAME = __name__.split('.')[-1]
ALL_OPTS = [
    template_file,
    driver,
    caching,
    cache_time,
    list_limit,
]


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=GROUP_NAME)


def list_opts():
    return {GROUP_NAME: ALL_OPTS}



# Copyright 2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import subprocess

import ldap.modlist
from six.moves import range
from six import PY2

from keystone.common import provider_api
import keystone.conf
from keystone import exception
from keystone.identity.backends import ldap as identity_ldap
from keystone.tests import unit
from keystone.tests.unit import test_backend_ldap


CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


def create_object(dn, attrs):
    if PY2:
        # NOTE: Once https://github.com/python-ldap/python-ldap/issues/249
        # is released, we can pass bytes_strictness='warn' as a parameter to
        # ldap.initialize instead of setting it after ldap.initialize.
        conn = ldap.initialize(CONF.ldap.url, bytes_mode=False)
        conn.bytes_strictness = 'warn'
    else:
        conn = ldap.initialize(CONF.ldap.url)
    conn.simple_bind_s(CONF.ldap.user, CONF.ldap.password)
    ldif = ldap.modlist.addModlist(attrs)
    conn.add_s(dn, ldif)
    conn.unbind_s()


class LiveLDAPIdentity(test_backend_ldap.LDAPIdentity):

    def setUp(self):
        self._ldap_skip_live()
        super(LiveLDAPIdentity, self).setUp()

    def _ldap_skip_live(self):
        self.skip_if_env_not_set('ENABLE_LDAP_LIVE_TEST')

    def clear_database(self):
        devnull = open('/dev/null', 'w')
        subprocess.call(['ldapdelete',
                         '-x',
                         '-D', CONF.ldap.user,
                         '-H', CONF.ldap.url,
                         '-w', CONF.ldap.password,
                         '-r', CONF.ldap.suffix],
                        stderr=devnull)

        if CONF.ldap.suffix.startswith('ou='):
            tree_dn_attrs = {'objectclass': 'organizationalUnit',
                             'ou': 'openstack'}
        else:
            tree_dn_attrs = {'objectclass': ['dcObject', 'organizationalUnit'],
                             'dc': 'openstack',
                             'ou': 'openstack'}
        create_object(CONF.ldap.suffix, tree_dn_attrs)
        create_object(CONF.ldap.user_tree_dn,
                      {'objectclass': 'organizationalUnit',
                       'ou': 'Users'})
        create_object(CONF.ldap.group_tree_dn,
                      {'objectclass': 'organizationalUnit',
                       'ou': 'UserGroups'})

    def config_files(self):
        config_files = super(LiveLDAPIdentity, self).config_files()
        config_files.append(unit.dirs.tests_conf('backend_liveldap.conf'))
        return config_files

    def test_build_tree(self):
        """Regression test for building the tree names."""
        # logic is different from the fake backend.
        user_api = identity_ldap.UserApi(CONF)
        self.assertTrue(user_api)
        self.assertEqual(CONF.ldap.user_tree_dn, user_api.tree_dn)

    def test_ldap_dereferencing(self):
        alt_users_ldif = {'objectclass': ['top', 'organizationalUnit'],
                          'ou': 'alt_users'}
        alt_fake_user_ldif = {'objectclass': ['person', 'inetOrgPerson'],
                              'cn': 'alt_fake1',
                              'sn': 'alt_fake1'}
        aliased_users_ldif = {'objectclass': ['alias', 'extensibleObject'],
                              'aliasedobjectname': "ou=alt_users,%s" %
                              CONF.ldap.suffix}
        create_object("ou=alt_users,%s" % CONF.ldap.suffix, alt_users_ldif)
        create_object("%s=alt_fake1,ou=alt_users,%s" %
                      (CONF.ldap.user_id_attribute, CONF.ldap.suffix),
                      alt_fake_user_ldif)
        create_object("ou=alt_users,%s" % CONF.ldap.user_tree_dn,
                      aliased_users_ldif)

        self.config_fixture.config(group='ldap',
                                   query_scope='sub',
                                   alias_dereferencing='never')
        PROVIDERS.identity_api = identity_ldap.Identity()
        self.assertRaises(exception.UserNotFound,
                          PROVIDERS.identity_api.get_user,
                          'alt_fake1')

        self.config_fixture.config(group='ldap',
                                   alias_dereferencing='searching')
        PROVIDERS.identity_api = identity_ldap.Identity()
        user_ref = PROVIDERS.identity_api.get_user('alt_fake1')
        self.assertEqual('alt_fake1', user_ref['id'])

        self.config_fixture.config(group='ldap', alias_dereferencing='always')
        PROVIDERS.identity_api = identity_ldap.Identity()
        user_ref = PROVIDERS.identity_api.get_user('alt_fake1')
        self.assertEqual('alt_fake1', user_ref['id'])

    # FakeLDAP does not correctly process filters, so this test can only be
    # run against a live LDAP server
    def test_list_groups_for_user_filtered(self):
        domain = self._get_domain_fixture()
        test_groups = []
        test_users = []
        GROUP_COUNT = 3
        USER_COUNT = 2

        positive_user = unit.create_user(PROVIDERS.identity_api, domain['id'])
        negative_user = unit.create_user(PROVIDERS.identity_api, domain['id'])

        for x in range(0, USER_COUNT):
            group_refs = PROVIDERS.identity_api.list_groups_for_user(
                test_users[x]['id'])
            self.assertEqual(0, len(group_refs))

        for x in range(0, GROUP_COUNT):
            new_group = unit.new_group_ref(domain_id=domain['id'])
            new_group = PROVIDERS.identity_api.create_group(new_group)
            test_groups.append(new_group)

            group_refs = PROVIDERS.identity_api.list_groups_for_user(
                positive_user['id'])
            self.assertEqual(x, len(group_refs))

            PROVIDERS.identity_api.add_user_to_group(
                positive_user['id'],
                new_group['id'])
            group_refs = PROVIDERS.identity_api.list_groups_for_user(
                positive_user['id'])
            self.assertEqual(x + 1, len(group_refs))

            group_refs = PROVIDERS.identity_api.list_groups_for_user(
                negative_user['id'])
            self.assertEqual(0, len(group_refs))

        driver = PROVIDERS.identity_api._select_identity_driver(
            CONF.identity.default_domain_id)
        driver.group.ldap_filter = '(dn=xx)'

        group_refs = PROVIDERS.identity_api.list_groups_for_user(
            positive_user['id'])
        self.assertEqual(0, len(group_refs))
        group_refs = PROVIDERS.identity_api.list_groups_for_user(
            negative_user['id'])
        self.assertEqual(0, len(group_refs))

        driver.group.ldap_filter = '(objectclass=*)'

        group_refs = PROVIDERS.identity_api.list_groups_for_user(
            positive_user['id'])
        self.assertEqual(GROUP_COUNT, len(group_refs))
        group_refs = PROVIDERS.identity_api.list_groups_for_user(
            negative_user['id'])
        self.assertEqual(0, len(group_refs))

    def test_user_enable_attribute_mask(self):
        self.config_fixture.config(
            group='ldap',
            user_enabled_emulation=False,
            user_enabled_attribute='employeeType')
        super(LiveLDAPIdentity, self).test_user_enable_attribute_mask()

    def test_create_project_case_sensitivity(self):
        # The attribute used for the live LDAP tests is case insensitive.

        def call_super():
            (super(LiveLDAPIdentity, self).
                test_create_project_case_sensitivity())

        self.assertRaises(exception.Conflict, call_super)

    def test_create_user_case_sensitivity(self):
        # The attribute used for the live LDAP tests is case insensitive.

        def call_super():
            super(LiveLDAPIdentity, self).test_create_user_case_sensitivity()

        self.assertRaises(exception.Conflict, call_super)

    def test_project_update_missing_attrs_with_a_falsey_value(self):
        # The description attribute doesn't allow an empty value.

        def call_super():
            (super(LiveLDAPIdentity, self).
                test_project_update_missing_attrs_with_a_falsey_value())

        self.assertRaises(ldap.INVALID_SYNTAX, call_super)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log
from oslo_serialization import msgpackutils
from oslo_utils import timeutils

from keystone.common import cache
from keystone.common import utils


LOG = log.getLogger(__name__)

# The set of attributes common between the RevokeEvent
# and the dictionaries created from the token Data.
_NAMES = ['trust_id',
          'consumer_id',
          'access_token_id',
          'audit_id',
          'audit_chain_id',
          'expires_at',
          'domain_id',
          'project_id',
          'user_id',
          'role_id']


# Additional arguments for creating a RevokeEvent
_EVENT_ARGS = ['issued_before', 'revoked_at']

# Names of attributes in the RevocationEvent, including "virtual" attributes.
# Virtual attributes are those added based on other values.
_EVENT_NAMES = _NAMES + ['domain_scope_id']

# Values that will be in the token data but not in the event.
# These will compared with event values that have different names.
# For example: both trustor_id and trustee_id are compared against user_id
_TOKEN_KEYS = ['identity_domain_id',
               'assignment_domain_id',
               'issued_at',
               'trustor_id',
               'trustee_id']

# Alternative names to be checked in token for every field in
# revoke tree.
ALTERNATIVES = {
    'user_id': ['user_id', 'trustor_id', 'trustee_id'],
    'domain_id': ['identity_domain_id', 'assignment_domain_id'],
    # For a domain-scoped token, the domain is in assignment_domain_id.
    'domain_scope_id': ['assignment_domain_id', ],
}


REVOKE_KEYS = _NAMES + _EVENT_ARGS


def blank_token_data(issued_at):
    token_data = dict()
    for name in _NAMES:
        token_data[name] = None
    for name in _TOKEN_KEYS:
        token_data[name] = None
    # required field
    token_data['issued_at'] = issued_at
    return token_data


class RevokeEvent(object):
    def __init__(self, **kwargs):
        for k in REVOKE_KEYS:
            v = kwargs.get(k)
            setattr(self, k, v)

        if self.domain_id and self.expires_at:
            # This is revoking a domain-scoped token.
            self.domain_scope_id = self.domain_id
            self.domain_id = None
        else:
            # This is revoking all tokens for a domain.
            self.domain_scope_id = None

        if self.expires_at is not None:
            # Trim off the expiration time because MySQL timestamps are only
            # accurate to the second.
            self.expires_at = self.expires_at.replace(microsecond=0)

        if self.revoked_at is None:
            self.revoked_at = timeutils.utcnow().replace(microsecond=0)
        if self.issued_before is None:
            self.issued_before = self.revoked_at

    def to_dict(self):
        keys = ['user_id',
                'role_id',
                'domain_id',
                'domain_scope_id',
                'project_id',
                'audit_id',
                'audit_chain_id',
                ]
        event = {key: self.__dict__[key] for key in keys
                 if self.__dict__[key] is not None}
        if self.trust_id is not None:
            event['OS-TRUST:trust_id'] = self.trust_id
        if self.consumer_id is not None:
            event['OS-OAUTH1:consumer_id'] = self.consumer_id
        if self.access_token_id is not None:
            event['OS-OAUTH1:access_token_id'] = self.access_token_id
        if self.expires_at is not None:
            event['expires_at'] = utils.isotime(self.expires_at)
        if self.issued_before is not None:
            event['issued_before'] = utils.isotime(self.issued_before,
                                                   subsecond=True)
        if self.revoked_at is not None:
            event['revoked_at'] = utils.isotime(self.revoked_at,
                                                subsecond=True)
        return event


def is_revoked(events, token_data):
    """Check if a token matches a revocation event.

    Compare a token against every revocation event. If the token matches an
    event in the `events` list, the token is revoked. If the token is compared
    against every item in the list without a match, it is not considered
    revoked from the `revoke_api`.

    :param events: a list of RevokeEvent instances
    :param token_data: map based on a flattened view of the token. The required
                       fields are `expires_at`,`user_id`, `project_id`,
                       `identity_domain_id`, `assignment_domain_id`,
                       `trust_id`, `trustor_id`, `trustee_id` `consumer_id` and
                       `access_token_id`
    :returns: True if the token matches an existing revocation event, meaning
              the token is revoked. False is returned if the token does not
              match any revocation events, meaning the token is considered
              valid by the revocation API.
    """
    return any([matches(e, token_data) for e in events])


def matches(event, token_values):
    """See if the token matches the revocation event.

    A brute force approach to checking.
    Compare each attribute from the event with the corresponding
    value from the token.  If the event does not have a value for
    the attribute, a match is still possible.  If the event has a
    value for the attribute, and it does not match the token, no match
    is possible, so skip the remaining checks.

    :param event: a RevokeEvent instance
    :param token_values: dictionary with set of values taken from the
                         token
    :returns: True if the token matches the revocation event, indicating the
              token has been revoked
    """
    # If any one check does not match, the whole token does
    # not match the event. The numerous return False indicate
    # that the token is still valid and short-circuits the
    # rest of the logic.

    # The token has two attributes that can match the domain_id.
    if event.domain_id is not None and event.domain_id not in(
            token_values['identity_domain_id'],
            token_values['assignment_domain_id'],):
        return False

    if event.domain_scope_id is not None and event.domain_scope_id not in (
            token_values['assignment_domain_id'],):
        return False

    # If an event specifies an attribute name, but it does not match, the token
    # is not revoked.
    if event.expires_at is not None and event.expires_at not in (
            token_values['expires_at'],):
        return False

    if event.trust_id is not None and event.trust_id not in (
            token_values['trust_id'],):
        return False

    if event.consumer_id is not None and event.consumer_id not in (
            token_values['consumer_id'],):
        return False

    if event.audit_chain_id is not None and event.audit_chain_id not in (
            token_values['audit_chain_id'],):
        return False

    if event.role_id is not None and event.role_id not in (
            token_values['roles']):
        return False

    return True


def build_token_values(token):

    token_expires_at = timeutils.parse_isotime(token.expires_at)

    # Trim off the microseconds because the revocation event only has
    # expirations accurate to the second.
    token_expires_at = token_expires_at.replace(microsecond=0)

    token_values = {
        'expires_at': timeutils.normalize_time(token_expires_at),
        'issued_at': timeutils.normalize_time(
            timeutils.parse_isotime(token.issued_at)),
        'audit_id': token.audit_id,
        'audit_chain_id': token.parent_audit_id,
    }

    if token.user_id is not None:
        token_values['user_id'] = token.user_id
        # Federated users do not have a domain, be defensive and get the user
        # domain set to None in the federated user case.
        token_values['identity_domain_id'] = token.user_domain['id']
    else:
        token_values['user_id'] = None
        token_values['identity_domain_id'] = None

    if token.project_id is not None:
        token_values['project_id'] = token.project_id
        # The domain_id of projects acting as domains is None
        token_values['assignment_domain_id'] = token.project_domain['id']
    else:
        token_values['project_id'] = None

    if token.domain_id is not None:
        token_values['assignment_domain_id'] = token.domain_id
    else:
        token_values['assignment_domain_id'] = None

    role_list = []
    if token.roles is not None:
        for role in token.roles:
            role_list.append(role['id'])
    token_values['roles'] = role_list

    if token.trust_scoped:
        token_values['trust_id'] = token.trust['id']
        token_values['trustor_id'] = token.trustor['id']
        token_values['trustee_id'] = token.trustee['id']
    else:
        token_values['trust_id'] = None
        token_values['trustor_id'] = None
        token_values['trustee_id'] = None

    if token.oauth_scoped:
        token_values['consumer_id'] = token.access_token['consumer_id']
        token_values['access_token_id'] = token.access_token['id']
    else:
        token_values['consumer_id'] = None
        token_values['access_token_id'] = None

    return token_values


class _RevokeEventHandler(object):
    # NOTE(morganfainberg): There needs to be reserved "registry" entries set
    # in oslo_serialization for application-specific handlers. We picked 127
    # here since it's waaaaaay far out before oslo_serialization will use it.
    identity = 127
    handles = (RevokeEvent,)

    def __init__(self, registry):
        self._registry = registry

    def serialize(self, obj):
        return msgpackutils.dumps(obj.__dict__, registry=self._registry)

    def deserialize(self, data):
        revoke_event_data = msgpackutils.loads(data, registry=self._registry)
        try:
            revoke_event = RevokeEvent(**revoke_event_data)
        except Exception:
            LOG.debug("Failed to deserialize RevokeEvent. Data is %s",
                      revoke_event_data)
            raise
        return revoke_event


cache.register_model_handler(_RevokeEventHandler)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import migrate
import sqlalchemy as sql

_USER_TABLE_NAME = 'user'
_USER_NAME_COLUMN_NAME = 'name'
_USER_DOMAINID_COLUMN_NAME = 'domain_id'
_USER_PASSWORD_COLUMN_NAME = 'password'  # nosec


def upgrade(migrate_engine):
    meta = sql.MetaData()
    meta.bind = migrate_engine

    user_table = sql.Table(_USER_TABLE_NAME, meta, autoload=True)

    # NOTE(gnuoy): the `domain_id` unique constraint is not guaranteed to
    # be a fixed name, such as 'ixu_user_name_domain_id`, so we need to
    # search for the correct constraint that only affects
    # user_table.c.domain_id and drop that constraint. (Fix based on
    # morganfainbergs fix in 088_domain_specific_roles.py)
    #
    # This is an idempotent change that reflects the fix to migration
    # 91 if the user name & domain_id unique constraint was not named
    # consistently and someone manually fixed the migrations / db
    # without dropping the old constraint.
    # This is a copy of migration 97 to catch any/all deployments that
    # are close to master. migration 97 will be backported to
    # stable/mitaka.

    to_drop = None
    if migrate_engine.name == 'mysql':
        for index in user_table.indexes:
            if (index.unique and len(index.columns) == 2 and
                    _USER_DOMAINID_COLUMN_NAME in index.columns and
                    _USER_NAME_COLUMN_NAME in index.columns):
                to_drop = index
                break
    else:
        for index in user_table.constraints:
            if (len(index.columns) == 2 and
                    _USER_DOMAINID_COLUMN_NAME in index.columns and
                    _USER_NAME_COLUMN_NAME in index.columns):
                to_drop = index
                break

    # remove domain_id and name unique constraint
    if to_drop is not None:
        migrate.UniqueConstraint(user_table.c.domain_id,
                                 user_table.c.name,
                                 name=to_drop.name).drop()

    # If migration 91 was aborted due to Bug #1572341 then columns may not
    # have been dropped.
    if _USER_DOMAINID_COLUMN_NAME in user_table.c:
        user_table.c.domain_id.drop()
    if _USER_NAME_COLUMN_NAME in user_table.c:
        user_table.c.name.drop()
    if _USER_PASSWORD_COLUMN_NAME in user_table.c:
        user_table.c.password.drop()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# This file handles all flask-restful resources for /v3/s3tokens

import base64
import hashlib
import hmac

import flask
from oslo_serialization import jsonutils
import six
from six.moves import http_client

from keystone.api._shared import EC2_S3_Resource
from keystone.api._shared import json_home_relations
from keystone.common import render_token
from keystone.common import utils
from keystone import exception
from keystone.i18n import _
from keystone.server import flask as ks_flask


def _calculate_signature_v1(string_to_sign, secret_key):
    """Calculate a v1 signature.

    :param bytes string_to_sign: String that contains request params and
                                 is used for calculate signature of request
    :param text secret_key: Second auth key of EC2 account that is used to
                            sign requests
    """
    key = str(secret_key).encode('utf-8')
    if six.PY2:
        b64_encode = base64.encodestring
    else:
        b64_encode = base64.encodebytes
    signed = b64_encode(hmac.new(key, string_to_sign, hashlib.sha1)
                        .digest()).decode('utf-8').strip()
    return signed


def _calculate_signature_v4(string_to_sign, secret_key):
    """Calculate a v4 signature.

    :param bytes string_to_sign: String that contains request params and
                                 is used for calculate signature of request
    :param text secret_key: Second auth key of EC2 account that is used to
                            sign requests
    """
    parts = string_to_sign.split(b'\n')
    if len(parts) != 4 or parts[0] != b'AWS4-HMAC-SHA256':
        raise exception.Unauthorized(message=_('Invalid EC2 signature.'))
    scope = parts[2].split(b'/')
    if len(scope) != 4 or scope[2] != b's3' or scope[3] != b'aws4_request':
        raise exception.Unauthorized(message=_('Invalid EC2 signature.'))

    def _sign(key, msg):
        return hmac.new(key, msg, hashlib.sha256).digest()

    signed = _sign(('AWS4' + secret_key).encode('utf-8'), scope[0])
    signed = _sign(signed, scope[1])
    signed = _sign(signed, scope[2])
    signed = _sign(signed, b'aws4_request')

    signature = hmac.new(signed, string_to_sign, hashlib.sha256)
    return signature.hexdigest()


class S3Resource(EC2_S3_Resource.ResourceBase):
    @staticmethod
    def _check_signature(creds_ref, credentials):
        string_to_sign = base64.urlsafe_b64decode(str(credentials['token']))

        if string_to_sign[0:4] != b'AWS4':
            signature = _calculate_signature_v1(string_to_sign,
                                                creds_ref['secret'])
        else:
            signature = _calculate_signature_v4(string_to_sign,
                                                creds_ref['secret'])

        if not utils.auth_str_equal(credentials['signature'], signature):
            raise exception.Unauthorized(
                message=_('Credential signature mismatch'))

    @ks_flask.unenforced_api
    def post(self):
        """Authenticate s3token.

        POST /v3/s3tokens
        """
        token = self.handle_authenticate()
        token_reference = render_token.render_token_response_from_model(token)
        resp_body = jsonutils.dumps(token_reference)
        response = flask.make_response(resp_body, http_client.OK)
        response.headers['Content-Type'] = 'application/json'
        return response


class S3Api(ks_flask.APIBase):
    _name = 's3tokens'
    _import_name = __name__
    resources = []
    resource_mapping = [
        ks_flask.construct_resource_map(
            resource=S3Resource,
            url='/s3tokens',
            resource_kwargs={},
            rel='s3tokens',
            resource_relation_func=(
                json_home_relations.s3_token_resource_rel_func))
    ]


APIs = (S3Api,)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sqlalchemy as sql

from keystone.credential.providers import fernet as credential_fernet


def upgrade(migrate_engine):
    meta = sql.MetaData()
    meta.bind = migrate_engine
    session = sql.orm.sessionmaker(bind=migrate_engine)()

    credential_table = sql.Table('credential', meta, autoload=True)
    credentials = list(credential_table.select().execute())

    for credential in credentials:
        crypto, keys = credential_fernet.get_multi_fernet_keys()
        primary_key_hash = credential_fernet.primary_key_hash(keys)
        encrypted_blob = crypto.encrypt(credential['blob'].encode('utf-8'))
        values = {
            'encrypted_blob': encrypted_blob,
            'key_hash': primary_key_hash
        }
        update = credential_table.update().where(
            credential_table.c.id == credential.id
        ).values(values)
        session.execute(update)
        session.commit()
    session.close()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystone.common import validation
from keystone.common.validation import parameter_types
from keystone.resource.backends import resource_options as ro

_name_properties = {
    'type': 'string',
    'minLength': 1,
    'maxLength': 64,
    'pattern': '[\S]+'
}

_project_tag_name_properties = {
    'type': 'string',
    'minLength': 1,
    'maxLength': 255,
    # NOTE(gagehugo) This pattern is for tags which follows the
    # guidelines as set by the API-WG, which matches anything that
    # does not contain a '/' or ','.
    # https://specs.openstack.org/openstack/api-wg/guidelines/tags.html
    'pattern': '^[^,/]*$'
}

_project_tags_list_properties = {
    'type': 'array',
    'items': _project_tag_name_properties,
    'required': [],
    'maxItems': 80,
    'uniqueItems': True
}

_project_properties = {
    'description': validation.nullable(parameter_types.description),
    # NOTE(htruta): domain_id is nullable for projects acting as a domain.
    'domain_id': validation.nullable(parameter_types.id_string),
    'enabled': parameter_types.boolean,
    'is_domain': parameter_types.boolean,
    'parent_id': validation.nullable(parameter_types.id_string),
    'name': _name_properties,
    'tags': _project_tags_list_properties,
    'options': ro.PROJECT_OPTIONS_REGISTRY.json_schema
}

# This is for updating a single project tag via the URL
project_tag_create = _project_tag_name_properties

# This is for updaing a project with a list of tags
project_tags_update = _project_tags_list_properties

project_create = {
    'type': 'object',
    'properties': _project_properties,
    # NOTE(lbragstad): A project name is the only parameter required for
    # project creation according to the Identity V3 API. We should think
    # about using the maxProperties validator here, and in update.
    'required': ['name'],
    'additionalProperties': True
}

project_update = {
    'type': 'object',
    'properties': _project_properties,
    # NOTE(lbragstad): Make sure at least one property is being updated
    'minProperties': 1,
    'additionalProperties': True
}

_domain_properties = {
    'description': validation.nullable(parameter_types.description),
    'enabled': parameter_types.boolean,
    'name': _name_properties,
    'tags': project_tags_update
}

domain_create = {
    'type': 'object',
    'properties': _domain_properties,
    # TODO(lbragstad): According to the V3 API spec, name isn't required but
    # the current implementation in assignment.controller:DomainV3 requires a
    # name for the domain.
    'required': ['name'],
    'additionalProperties': True
}

domain_update = {
    'type': 'object',
    'properties': _domain_properties,
    'minProperties': 1,
    'additionalProperties': True
}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import versionutils
from oslo_policy import policy

from keystone.common.policies import base

deprecated_get_implied_role = policy.DeprecatedRule(
    name=base.IDENTITY % 'get_implied_role',
    check_str=base.RULE_ADMIN_REQUIRED
)
deprecated_list_implied_roles = policy.DeprecatedRule(
    name=base.IDENTITY % 'list_implied_roles',
    check_str=base.RULE_ADMIN_REQUIRED,
)
deprecated_list_role_inference_rules = policy.DeprecatedRule(
    name=base.IDENTITY % 'list_role_inference_rules',
    check_str=base.RULE_ADMIN_REQUIRED,
)
deprecated_check_implied_role = policy.DeprecatedRule(
    name=base.IDENTITY % 'check_implied_role',
    check_str=base.RULE_ADMIN_REQUIRED,
)
deprecated_create_implied_role = policy.DeprecatedRule(
    name=base.IDENTITY % 'create_implied_role',
    check_str=base.RULE_ADMIN_REQUIRED,
)
deprecated_delete_implied_role = policy.DeprecatedRule(
    name=base.IDENTITY % 'delete_implied_role',
    check_str=base.RULE_ADMIN_REQUIRED,
)

DEPRECATED_REASON = (
    "The implied role API is now aware of system scope and default roles."
)


implied_role_policies = [
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_implied_role',
        check_str=base.SYSTEM_READER,
        # FIXME(lbragstad) The management of implied roles currently makes
        # sense as a system-only resource. Once keystone has the ability to
        # support RBAC solely over the API without having to customize policy
        # files, scope_types should include 'project'.
        scope_types=['system'],
        description='Get information about an association between two roles. '
                    'When a relationship exists between a prior role and an '
                    'implied role and the prior role is assigned to a user, '
                    'the user also assumes the implied role.',
        operations=[
            {'path': '/v3/roles/{prior_role_id}/implies/{implied_role_id}',
             'method': 'GET'}],
        deprecated_rule=deprecated_get_implied_role,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.TRAIN),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_implied_roles',
        check_str=base.SYSTEM_READER,
        scope_types=['system'],
        description='List associations between two roles. When a relationship '
                    'exists between a prior role and an implied role and the '
                    'prior role is assigned to a user, the user also assumes '
                    'the implied role. This will return all the implied roles '
                    'that would be assumed by the user who gets the specified '
                    'prior role.',
        operations=[
            {'path': '/v3/roles/{prior_role_id}/implies', 'method': 'GET'},
            {'path': '/v3/roles/{prior_role_id}/implies', 'method': 'HEAD'}],
        deprecated_rule=deprecated_list_implied_roles,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.TRAIN),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'create_implied_role',
        check_str=base.SYSTEM_ADMIN,
        scope_types=['system'],
        description='Create an association between two roles. When a '
                    'relationship exists between a prior role and an implied '
                    'role and the prior role is assigned to a user, the user '
                    'also assumes the implied role.',
        operations=[
            {'path': '/v3/roles/{prior_role_id}/implies/{implied_role_id}',
             'method': 'PUT'}],
        deprecated_rule=deprecated_create_implied_role,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.TRAIN),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'delete_implied_role',
        check_str=base.SYSTEM_ADMIN,
        scope_types=['system'],
        description='Delete the association between two roles. When a '
                    'relationship exists between a prior role and an implied '
                    'role and the prior role is assigned to a user, the user '
                    'also assumes the implied role. Removing the association '
                    'will cause that effect to be eliminated.',
        operations=[
            {'path': '/v3/roles/{prior_role_id}/implies/{implied_role_id}',
             'method': 'DELETE'}],
        deprecated_rule=deprecated_delete_implied_role,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.TRAIN),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_role_inference_rules',
        check_str=base.SYSTEM_READER,
        scope_types=['system'],
        description='List all associations between two roles in the system. '
                    'When a relationship exists between a prior role and an '
                    'implied role and the prior role is assigned to a user, '
                    'the user also assumes the implied role.',
        operations=[
            {'path': '/v3/role_inferences', 'method': 'GET'},
            {'path': '/v3/role_inferences', 'method': 'HEAD'}],
        deprecated_rule=deprecated_list_role_inference_rules,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.TRAIN),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'check_implied_role',
        check_str=base.SYSTEM_READER,
        scope_types=['system'],
        description='Check an association between two roles. When a '
                    'relationship exists between a prior role and an implied '
                    'role and the prior role is assigned to a user, the user '
                    'also assumes the implied role.',
        operations=[
            {'path': '/v3/roles/{prior_role_id}/implies/{implied_role_id}',
             'method': 'HEAD'}],
        deprecated_rule=deprecated_check_implied_role,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.TRAIN),
]


def list_rules():
    return implied_role_policies



# Copyright 2018 Huawei
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystone.common import validation
from keystone.common.validation import parameter_types

_registered_limit_properties = {
    'service_id': parameter_types.id_string,
    'region_id': {
        'type': ['null', 'string']
    },
    'resource_name': {
        'type': 'string',
        'minLength': 1,
        'maxLength': 255
    },
    'default_limit': {
        'type': 'integer',
        'minimum': -1,
        'maximum': 0x7FFFFFFF  # The maximum value a signed INT may have
    },
    'description': validation.nullable(parameter_types.description)
}

_registered_limit_create = {
    'type': 'object',
    'properties': _registered_limit_properties,
    'additionalProperties': False,
    'required': ['service_id', 'resource_name', 'default_limit']
}

registered_limit_create = {
    'type': 'array',
    'items': _registered_limit_create,
    'minItems': 1
}
registered_limit_update = {
    'type': 'object',
    'properties': _registered_limit_properties,
    'additionalProperties': False,
}

_project_limit_create_properties = {
    'project_id': parameter_types.id_string,
    'service_id': parameter_types.id_string,
    'region_id': {
        'type': 'string'
    },
    'resource_name': {
        'type': 'string',
        'minLength': 1,
        'maxLength': 255
    },
    'resource_limit': {
        'type': 'integer',
        'minimum': -1,
        'maximum': 0x7FFFFFFF  # The maximum value a signed INT may have
    },
    'description': validation.nullable(parameter_types.description)
}

_domain_limit_create_properties = {
    'domain_id': parameter_types.id_string,
    'service_id': parameter_types.id_string,
    'region_id': {
        'type': 'string'
    },
    'resource_name': {
        'type': 'string',
        'minLength': 1,
        'maxLength': 255
    },
    'resource_limit': {
        'type': 'integer',
        'minimum': -1,
        'maximum': 0x7FFFFFFF  # The maximum value a signed INT may have
    },
    'description': validation.nullable(parameter_types.description)
}

_limit_create = {
    'type': 'object',
    'oneOf': [
        {'properties': _project_limit_create_properties,
         'required': ['project_id', 'service_id', 'resource_name',
                      'resource_limit'],
         'additionalProperties': False,
         },
        {'properties': _domain_limit_create_properties,
         'required': ['domain_id', 'service_id', 'resource_name',
                      'resource_limit'],
         'additionalProperties': False,
         },
    ]
}

limit_create = {
    'type': 'array',
    'items': _limit_create,
    'minItems': 1
}

_limit_update_properties = {
    'resource_limit': {
        'type': 'integer',
        'minimum': -1,
        'maximum': 0x7FFFFFFF  # The maximum value a signed INT may have
    },
    'description': validation.nullable(parameter_types.description)
}

limit_update = {
    'type': 'object',
    'properties': _limit_update_properties,
    'additionalProperties': False
}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import collections
import functools
import itertools
import re
import uuid
import wsgiref.util

import flask
from flask import blueprints
import flask_restful
import flask_restful.utils
from oslo_log import log
from oslo_log import versionutils
from oslo_serialization import jsonutils
import six
from six.moves import http_client

from keystone.common import authorization
from keystone.common import context
from keystone.common import driver_hints
from keystone.common import json_home
from keystone.common.rbac_enforcer import enforcer
from keystone.common import utils
import keystone.conf
from keystone import exception
from keystone.i18n import _
from keystone import notifications


# NOTE(morgan): Capture the relevant part of the flask url route rule for
# substitution. In flask arguments (e.g. url elements to be passed to the
# "resource" method, e.g. user_id, are specified like `<string:user_id>`
# we use this regex to replace the <> with {} for JSON Home purposes and
# remove the argument type. Use of this is done like
# _URL_SUBST.sub('{\\1}', entity_path), which replaces the whole match
# match rule bit with the capture group (this is a greedy sub).
_URL_SUBST = re.compile(r'<[^\s:]+:([^>]+)>')
CONF = keystone.conf.CONF
LOG = log.getLogger(__name__)
ResourceMap = collections.namedtuple(
    'resource_map', 'resource, url, alternate_urls, kwargs, json_home_data')
JsonHomeData = collections.namedtuple(
    'json_home_data', 'rel, status, path_vars')

_v3_resource_relation = json_home.build_v3_resource_relation


def construct_resource_map(resource, url, resource_kwargs, alternate_urls=None,
                           rel=None, status=json_home.Status.STABLE,
                           path_vars=None,
                           resource_relation_func=_v3_resource_relation):
    """Construct the ResourceMap Named Tuple.

    :param resource: The flask-RESTful resource class implementing the methods
                     for the API.
    :type resource: :class:`ResourceMap`
    :param url: Flask-standard url route, all flask url routing rules apply.
                url variables will be passed to the Resource methods as
                arguments.
    :type url: str
    :param resource_kwargs: a dict of optional value(s) that can further modify
                            the handling of the routing.

                            * endpoint: endpoint name (defaults to
                                        :meth:`Resource.__name__.lower`
                                        Can be used to reference this route in
                                        :class:`fields.Url` fields (str)

                            * resource_class_args: args to be forwarded to the
                                                   constructor of the resource.
                                                   (tuple)

                            * resource_class_kwargs: kwargs to be forwarded to
                                                     the constructor of the
                                                     resource. (dict)

                            Additional keyword arguments not specified above
                            will be passed as-is to
                            :meth:`flask.Flask.add_url_rule`.
    :param alternate_urls: An iterable (list) of dictionaries containing urls
                           and associated json home REL data. Each element is
                           expected to be a dictionary with a 'url' key and an
                           optional 'json_home' key for a 'JsonHomeData' named
                           tuple  These urls will also map to the resource.
                           These are used to ensure API compatibility when a
                           "new" path is more correct for the API but old paths
                           must continue to work. Example:
                           `/auth/domains` being the new path for
                           `/OS-FEDERATION/domains`. The `OS-FEDERATION` part
                           would be listed as an alternate url. If a
                           'json_home' key is provided, the original path
                           with the new json_home data will be added to the
                           JSON Home Document.
    :type: iterable or None
    :param rel:
    :type rel: str or None
    :param status: JSON Home API Status, e.g. "STABLE"
    :type status: str
    :param path_vars: JSON Home Path Var Data (arguments)
    :type path_vars: dict or None
    :param resource_relation_func: function to build expected resource rel data
    :type resource_relation_func: callable
    :return:
    """
    if rel is not None:
        jh_data = construct_json_home_data(
            rel=rel, status=status, path_vars=path_vars,
            resource_relation_func=resource_relation_func)
    else:
        jh_data = None
    if not url.startswith('/'):
        url = '/%s' % url
    return ResourceMap(
        resource=resource, url=url, alternate_urls=alternate_urls,
        kwargs=resource_kwargs, json_home_data=jh_data)


def construct_json_home_data(rel, status=json_home.Status.STABLE,
                             path_vars=None,
                             resource_relation_func=_v3_resource_relation):
    rel = resource_relation_func(resource_name=rel)
    return JsonHomeData(rel=rel, status=status, path_vars=(path_vars or {}))


def _initialize_rbac_enforcement_check():
    setattr(flask.g, enforcer._ENFORCEMENT_CHECK_ATTR, False)


def _assert_rbac_enforcement_called(resp):
    # assert is intended to be used to ensure code during development works
    # as expected, it is fine to be optimized out with `python -O`
    msg = ('PROGRAMMING ERROR: enforcement (`keystone.common.rbac_enforcer.'
           'enforcer.RBACEnforcer.enforce_call()`) has not been called; API '
           'is unenforced.')
    g = flask.g
    # NOTE(morgan): OPTIONS is a special case and is handled by flask
    # internally. We should never be enforcing on OPTIONS calls.
    if flask.request.method != 'OPTIONS':
        assert getattr(  # nosec
            g, enforcer._ENFORCEMENT_CHECK_ATTR, False), msg  # nosec
    return resp


def _remove_content_type_on_204(resp):
    # Remove content-type if the resp is 204.
    if resp.status_code == http_client.NO_CONTENT:
        resp.headers.pop('content-type', None)
    return resp


@six.add_metaclass(abc.ABCMeta)
class APIBase(object):

    @property
    @abc.abstractmethod
    def _name(self):
        """Override with an attr consisting of the API Name, e.g 'users'."""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def _import_name(self):
        """Override with an attr consisting of the value of `__name__`."""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def resource_mapping(self):
        """An attr containing of an iterable of :class:`ResourceMap`.

        Each :class:`ResourceMap` is a NamedTuple with the following elements:

            * resource: a :class:`flask_restful.Resource` class or subclass

            * url: a url route to match for the resource, standard flask
                   routing rules apply. Any url variables will be passed
                   to the resource method as args. (str)

            * alternate_urls: an iterable of url routes to match for the
                              resource, standard flask routing rules apply.
                              These rules are in addition (for API compat) to
                              the primary url. Any url variables will be
                              passed to the resource method as args. (iterable)

            * json_home_data: :class:`JsonHomeData` populated with relevant
                              info for populated JSON Home Documents or None.

            * kwargs: a dict of optional value(s) that can further modify the
                      handling of the routing.

                      * endpoint: endpoint name (defaults to
                                  :meth:`Resource.__name__.lower`
                                  Can be used to reference this route in
                                  :class:`fields.Url` fields (str)

                      * resource_class_args: args to be forwarded to the
                                             constructor of the resource.
                                             (tuple)

                      * resource_class_kwargs: kwargs to be forwarded to the
                                               constructor of the resource.
                                               (dict)

                      Additional keyword arguments not specified above will be
                      passed as-is to :meth:`flask.Flask.add_url_rule`.
        """
        raise NotImplementedError()

    @property
    def resources(self):
        return []

    @staticmethod
    def _build_bp_url_prefix(prefix):
        # NOTE(morgan): Keystone only has a V3 API, this is here for future
        # proofing and exceptional cases such as root discovery API object(s)
        parts = ['/v3']
        if prefix:
            parts.append(prefix.lstrip('/'))
        return '/'.join(parts).rstrip('/')

    @property
    def api(self):
        # The API may be directly accessed via this property
        return self.__api

    @property
    def blueprint(self):
        # The API Blueprint may be directly accessed via this property
        return self.__blueprint

    def __init__(self, blueprint_url_prefix='', api_url_prefix='',
                 default_mediatype='application/json', decorators=None,
                 errors=None):
        self.__before_request_functions_added = False
        self.__after_request_functions_added = False

        self._default_mediatype = default_mediatype
        blueprint_url_prefix = blueprint_url_prefix.rstrip('/')
        api_url_prefix = api_url_prefix.rstrip('/')

        if api_url_prefix and not api_url_prefix.startswith('/'):
            self._api_url_prefix = '/%s' % api_url_prefix
        else:
            # NOTE(morgan): If the api_url_prefix is empty fall back on the
            # class-level defined `_api_url_prefix` if it is set.
            self._api_url_prefix = (api_url_prefix or
                                    getattr(self, '_api_url_prefix', ''))

        if blueprint_url_prefix and not blueprint_url_prefix.startswith('/'):
            self._blueprint_url_prefix = self._build_bp_url_prefix(
                '/%s' % blueprint_url_prefix)
        else:
            self._blueprint_url_prefix = self._build_bp_url_prefix(
                blueprint_url_prefix)

        self.__blueprint = blueprints.Blueprint(
            name=self._name, import_name=self._import_name,
            url_prefix=self._blueprint_url_prefix)
        self.__api = flask_restful.Api(
            app=self.__blueprint, prefix=self._api_url_prefix,
            default_mediatype=self._default_mediatype,
            decorators=decorators, errors=errors)

        # NOTE(morgan): Make sure we're using oslo_serialization.jsonutils
        # instead of the default json serializer. Keystone has data types that
        # the default serializer cannot handle, representation is a decorator
        # but since we instantiate the API in-line we need to do some magic
        # and call it as a normal method.
        self.__api.representation('application/json')(self._output_json)

        self._add_resources()
        self._add_mapped_resources()

        # Apply Before and After request functions
        self._register_before_request_functions()
        self._register_after_request_functions()
        # Assert is intended to ensure code works as expected in development,
        # it is fine to optimize out with python -O
        msg = '%s_request functions not registered'
        assert self.__before_request_functions_added, msg % 'before'  # nosec
        assert self.__after_request_functions_added, msg % 'after'  # nosec

    def _add_resources(self):
        # Add resources that are standardized. Each resource implements a
        # base set of handling for a collection of entities such as
        # `users`. Resources are sourced from self.resources. Each resource
        # should have an attribute/property containing the `collection_key`
        # which is typically the "plural" form of the entity, e.g. `users` and
        # `member_key` which is typically the "singular" of the entity, e.g.
        # `user`. Resources are sourced from self.resources, each element is
        # simply a :class:`flask_restful.Resource`.
        for r in self.resources:
            c_key = getattr(r, 'collection_key', None)
            m_key = getattr(r, 'member_key', None)
            r_pfx = getattr(r, 'api_prefix', None)

            if not c_key or not m_key:
                LOG.debug('Unable to add resource %(resource)s to API '
                          '%(name)s, both `member_key` and `collection_key` '
                          'must be implemented. [collection_key(%(col_key)s) '
                          'member_key(%(m_key)s)]',
                          {'resource': r.__name__,
                           'name': self._name, 'col_key': c_key,
                           'm_key': m_key})
                continue
            if r_pfx != self._api_url_prefix:
                LOG.debug('Unable to add resource %(resource)s to API as the '
                          'API Prefixes do not match: %(apfx)r != %(rpfx)r',
                          {'resource': r.__name__,
                           'rpfx': r_pfx, 'apfx': self._api_url_prefix})
                continue

            # NOTE(morgan): The Prefix is automatically added by the API, so
            # we do not add it to the paths here.
            collection_path = '/%s' % c_key
            if getattr(r, '_id_path_param_name_override', None):
                # The member_key doesn't match the "id" key in the url, make
                # sure to use the correct path-key for ID.
                member_id_key = getattr(r, '_id_path_param_name_override')
            else:
                member_id_key = '%(member_key)s_id' % {'member_key': m_key}

            entity_path = '/%(collection)s/<string:%(member)s>' % {
                'collection': c_key, 'member': member_id_key}
            # NOTE(morgan): The json-home form of the entity path is different
            # from the flask-url routing form. Must also include the prefix
            jh_e_path = _URL_SUBST.sub('{\\1}', '%(pfx)s/%(e_path)s' % {
                'pfx': self._api_url_prefix,
                'e_path': entity_path.lstrip('/')})

            LOG.debug(
                'Adding standard routes to API %(name)s for `%(resource)s` '
                '(API Prefix: %(prefix)s) [%(collection_path)s, '
                '%(entity_path)s]', {
                    'name': self._name, 'resource': r.__class__.__name__,
                    'collection_path': collection_path,
                    'entity_path': entity_path,
                    'prefix': self._api_url_prefix})
            self.api.add_resource(r, collection_path, entity_path)

            # Add JSON Home data
            resource_rel_func = getattr(
                r, 'json_home_resource_rel_func',
                json_home.build_v3_resource_relation)
            resource_rel_status = getattr(
                r, 'json_home_resource_status', None)
            collection_rel_resource_name = getattr(
                r, 'json_home_collection_resource_name_override', c_key)
            collection_rel = resource_rel_func(
                resource_name=collection_rel_resource_name)
            # NOTE(morgan): Add the prefix explicitly for JSON Home documents
            # to the collection path.
            href_val = '%(pfx)s%(collection_path)s' % {
                'pfx': self._api_url_prefix,
                'collection_path': collection_path}

            # If additional parameters exist in the URL, add them to the
            # href-vars dict.
            additional_params = getattr(
                r, 'json_home_additional_parameters', {})

            if additional_params:
                # NOTE(morgan): Special case, we have 'additional params' which
                # means we know the params are in the "prefix". This guarantees
                # the correct data in the json_home document with href-template
                # and href-vars even on the "collection" entry
                rel_data = dict()
                rel_data['href-template'] = _URL_SUBST.sub('{\\1}', href_val)
                rel_data['href-vars'] = additional_params
            else:
                rel_data = {'href': href_val}
            member_rel_resource_name = getattr(
                r, 'json_home_member_resource_name_override', m_key)

            entity_rel = resource_rel_func(
                resource_name=member_rel_resource_name)
            id_str = member_id_key

            parameter_rel_func = getattr(
                r, 'json_home_parameter_rel_func',
                json_home.build_v3_parameter_relation)
            id_param_rel = parameter_rel_func(parameter_name=id_str)
            entity_rel_data = {'href-template': jh_e_path,
                               'href-vars': {id_str: id_param_rel}}

            if additional_params:
                entity_rel_data.setdefault('href-vars', {}).update(
                    additional_params)

            if resource_rel_status is not None:
                json_home.Status.update_resource_data(
                    rel_data, resource_rel_status)
                json_home.Status.update_resource_data(
                    entity_rel_data, resource_rel_status)

            json_home.JsonHomeResources.append_resource(
                collection_rel, rel_data)
            json_home.JsonHomeResources.append_resource(
                entity_rel, entity_rel_data)

    def _add_mapped_resources(self):
        # Add resource mappings, non-standard resource connections
        for r in self.resource_mapping:
            alt_url_json_home_data = []
            LOG.debug(
                'Adding resource routes to API %(name)s: '
                '[%(url)r %(kwargs)r]',
                {'name': self._name, 'url': r.url, 'kwargs': r.kwargs})
            urls = [r.url]
            if r.alternate_urls is not None:
                for element in r.alternate_urls:
                    if self._api_url_prefix:
                        LOG.debug(
                            'Unable to add additional resource route '
                            '`%(route)s` to API %(name)s because API has a '
                            'URL prefix. Only APIs without explicit prefixes '
                            'can have alternate URL routes added.',
                            {'route': element['url'], 'name': self._name}
                        )
                        continue
                    LOG.debug(
                        'Adding additional resource route (alternate) to API '
                        '%(name)s: [%(url)r %(kwargs)r]',
                        {'name': self._name, 'url': element['url'],
                         'kwargs': r.kwargs})
                    urls.append(element['url'])
                    if element.get('json_home'):
                        alt_url_json_home_data.append(element['json_home'])
            # Add all URL routes at once.
            self.api.add_resource(r.resource, *urls, **r.kwargs)

            # Build the JSON Home data and add it to the relevant JSON Home
            # Documents for explicit JSON Home data.
            if r.json_home_data:
                resource_data = {}
                # NOTE(morgan): JSON Home form of the URL is different
                # from FLASK, do the conversion here.
                conv_url = '%(pfx)s/%(url)s' % {
                    'url': _URL_SUBST.sub('{\\1}', r.url).lstrip('/'),
                    'pfx': self._api_url_prefix}

                if r.json_home_data.path_vars:
                    resource_data['href-template'] = conv_url
                    resource_data['href-vars'] = r.json_home_data.path_vars
                else:
                    resource_data['href'] = conv_url
                json_home.Status.update_resource_data(
                    resource_data, r.json_home_data.status)
                json_home.JsonHomeResources.append_resource(
                    r.json_home_data.rel,
                    resource_data)

                for element in alt_url_json_home_data:
                    # Append the "new" path (resource) data with the old rel
                    # reference.
                    json_home.JsonHomeResources.append_resource(
                        element.rel, resource_data)

    def _register_before_request_functions(self, functions=None):
        """Register functions to be executed in the `before request` phase.

        Override this method and pass in via "super" any additional functions
        that should be registered. It is assumed that any override will also
        accept a "functions" list and append the passed in values to it's
        list prior to calling super.

        Each function will be called with no arguments and expects a NoneType
        return. If the function returns a value, that value will be returned
        as the response to the entire request, no further processing will
        happen.

        :param functions: list of functions that will be run in the
                          `before_request` phase.
        :type functions: list
        """
        functions = functions or []
        # Assert is intended to ensure code works as expected in development,
        # it is fine to optimize out with python -O
        msg = 'before_request functions already registered'
        assert not self.__before_request_functions_added, msg  # nosec
        # register global before request functions
        # e.g. self.__blueprint.before_request(function)
        self.__blueprint.before_request(_initialize_rbac_enforcement_check)

        # Add passed-in functions
        for f in functions:
            self.__blueprint.before_request(f)
        self.__before_request_functions_added = True

    def _register_after_request_functions(self, functions=None):
        """Register functions to be executed in the `after request` phase.

        Override this method and pass in via "super" any additional functions
        that should be registered. It is assumed that any override will also
        accept a "functions" list and append the passed in values to it's
        list prior to calling super.

        Each function will be called with a single argument of the Response
        class type. The function must return either the passed in Response or
        a new Response. NOTE: As of flask 0.7, these functions may not be
        executed in the case of an unhandled exception.

        :param functions: list of functions that will be run in the
                          `after_request` phase.
        :type functions: list
        """
        functions = functions or []
        # Assert is intended to ensure code works as expected in development,
        # it is fine to optimize out with python -O
        msg = 'after_request functions already registered'
        assert not self.__after_request_functions_added, msg  # nosec
        # register global after request functions
        # e.g. self.__blueprint.after_request(function)
        self.__blueprint.after_request(_assert_rbac_enforcement_called)
        self.__blueprint.after_request(_remove_content_type_on_204)

        # Add Passed-In Functions
        for f in functions:
            self.__blueprint.after_request(f)
        self.__after_request_functions_added = True

    @staticmethod
    def _output_json(data, code, headers=None):
        """Make a Flask response with a JSON encoded body.

        This is a replacement of the default that is shipped with flask-RESTful
        as we need oslo_serialization for the wider datatypes in our objects
        that are serialized to json.
        """
        settings = flask.current_app.config.get('RESTFUL_JSON', {})

        # If we're in debug mode, and the indent is not set, we set it to
        # a reasonable value here. Note that this won't override any existing
        # value that was set. We also set the "sort_keys" value.
        if flask.current_app.debug:
            settings.setdefault('indent', 4)
            settings.setdefault('sort_keys', not flask_restful.utils.PY3)

        # always end the json dumps with a new line
        # see https://github.com/mitsuhiko/flask/pull/1262
        dumped = jsonutils.dumps(data, **settings) + "\n"

        resp = flask.make_response(dumped, code)
        resp.headers.extend(headers or {})
        return resp

    @classmethod
    def instantiate_and_register_to_app(cls, flask_app):
        """Build the API object and register to the passed in flask_app.

        This is a simplistic loader that makes assumptions about how the
        blueprint is loaded. Anything beyond defaults should be done
        explicitly via normal instantiation where more values may be passed
        via :meth:`__init__`.

        :returns: :class:`keystone.server.flask.common.APIBase`
        """
        inst = cls()
        flask_app.register_blueprint(inst.blueprint)
        return inst


class _AttributeRaisesError(object):
    # NOTE(morgan): This is a special case class that exists to effectively
    # create a @classproperty style function. We use __get__ to raise the
    # exception.

    def __init__(self, name):
        self.__msg = 'PROGRAMMING ERROR: `self.{name}` is not set.'.format(
            name=name)

    def __get__(self, instance, owner):
        raise ValueError(self.__msg)


class ResourceBase(flask_restful.Resource):
    collection_key = _AttributeRaisesError(name='collection_key')
    member_key = _AttributeRaisesError(name='member_key')
    _public_parameters = frozenset([])
    # NOTE(morgan): This must match the string on the API the resource is
    # registered to.
    api_prefix = ''
    _id_path_param_name_override = None

    method_decorators = []

    @staticmethod
    def _assign_unique_id(ref):
        ref = ref.copy()
        ref['id'] = uuid.uuid4().hex
        return ref

    @staticmethod
    def _validate_id_format(id):
        uval = uuid.UUID(id).hex
        if uval != id:
            raise ValueError('badly formed hexadecimal UUID value')

    @classmethod
    def _require_matching_id(cls, ref):
        """Ensure the value matches the reference's ID, if any."""
        id_arg = None
        if cls.member_key is not None:
            id_arg = flask.request.view_args.get('%s_id' % cls.member_key)
        if ref.get('id') is not None and id_arg != ref['id']:
            raise exception.ValidationError('Cannot change ID')

    @classmethod
    def filter_params(cls, ref):
        """Remove unspecified parameters from the dictionary.

        This function removes unspecified parameters from the dictionary.
        This method checks only root-level keys from a ref dictionary.

        :param ref: a dictionary representing deserialized response to be
                    serialized
        """
        # NOTE(morgan): if _public_parameters is empty, do nothing. We do not
        # filter if we do not have an explicit white-list to work from.
        if cls._public_parameters:
            ref_keys = set(ref.keys())
            blocked_keys = ref_keys - cls._public_parameters
            for blocked_param in blocked_keys:
                del ref[blocked_param]
        return ref

    @classmethod
    def wrap_collection(cls, refs, hints=None, collection_name=None):
        """Wrap a collection, checking for filtering and pagination.

        Returns the wrapped collection, which includes:
        - Executing any filtering not already carried out
        - Truncate to a set limit if necessary
        - Adds 'self' links in every member
        - Adds 'next', 'self' and 'prev' links for the whole collection.

        :param refs: the list of members of the collection
        :param hints: list hints, containing any relevant filters and limit.
                      Any filters already satisfied by managers will have been
                      removed
        :param collection_name: optional override for the 'collection key'
                                class attribute. This is to be used when
                                wrapping a collection for a different api,
                                e.g. 'roles' from the 'trust' api.
        """
        # Check if there are any filters in hints that were not handled by
        # the drivers. The driver will not have paginated or limited the
        # output if it found there were filters it was unable to handle

        if hints:
            refs = cls.filter_by_attributes(refs, hints)

        list_limited, refs = cls.limit(refs, hints)

        collection = collection_name or cls.collection_key

        for ref in refs:
            cls._add_self_referential_link(ref, collection_name=collection)

        container = {collection: refs}
        self_url = full_url(flask.request.environ['PATH_INFO'])
        container['links'] = {
            'next': None,
            'self': self_url,
            'previous': None
        }
        if list_limited:
            container['truncated'] = True

        return container

    @classmethod
    def wrap_member(cls, ref, collection_name=None, member_name=None):
        cls._add_self_referential_link(ref, collection_name)
        return {member_name or cls.member_key: ref}

    @classmethod
    def _add_self_referential_link(cls, ref, collection_name=None):
        collection_element = collection_name or cls.collection_key
        if cls.api_prefix:
            api_prefix = cls.api_prefix.lstrip('/').rstrip('/')
            # ensure we have substituted the flask-arg specification
            # to the "keystone" mechanism, then format the string
            api_prefix = _URL_SUBST.sub('{\\1}', api_prefix)
            if flask.request.view_args:
                # if a prefix has substitutions it is *required* that the
                # values are passed as view_args to the HTTP action method
                # (e.g. head/get/post/...).
                api_prefix = api_prefix.format(**flask.request.view_args)
            collection_element = '/'.join(
                [api_prefix, collection_name or cls.collection_key])
        self_link = base_url(path='/'.join([collection_element, ref['id']]))
        ref.setdefault('links', {})['self'] = self_link

    @classmethod
    def filter_by_attributes(cls, refs, hints):
        """Filter a list of references by filter values."""
        def _attr_match(ref_attr, val_attr):
            """Matche attributes allowing for booleans as strings.

            We test explicitly for a value that defines it as 'False',
            which also means that the existence of the attribute with
            no value implies 'True'

            """
            if type(ref_attr) is bool:
                return ref_attr == utils.attr_as_boolean(val_attr)
            else:
                return ref_attr == val_attr

        def _inexact_attr_match(inexact_filter, ref):
            """Apply an inexact filter to a result dict.

            :param inexact_filter: the filter in question
            :param ref: the dict to check

            :returns: True if there is a match

            """
            comparator = inexact_filter['comparator']
            key = inexact_filter['name']

            if key in ref:
                filter_value = inexact_filter['value']
                target_value = ref[key]
                if not inexact_filter['case_sensitive']:
                    # We only support inexact filters on strings so
                    # it's OK to use lower()
                    filter_value = filter_value.lower()
                    target_value = target_value.lower()

                if comparator == 'contains':
                    return (filter_value in target_value)
                elif comparator == 'startswith':
                    return target_value.startswith(filter_value)
                elif comparator == 'endswith':
                    return target_value.endswith(filter_value)
                else:
                    # We silently ignore unsupported filters
                    return True

            return False

        for f in hints.filters:
            if f['comparator'] == 'equals':
                attr = f['name']
                value = f['value']
                refs = [r for r in refs if _attr_match(
                    utils.flatten_dict(r).get(attr), value)]
            else:
                # It might be an inexact filter
                refs = [r for r in refs if _inexact_attr_match(f, r)]

        return refs

    @property
    def auth_context(self):
        return flask.request.environ.get(authorization.AUTH_CONTEXT_ENV, None)

    @property
    def oslo_context(self):
        return flask.request.environ.get(context.REQUEST_CONTEXT_ENV, None)

    @property
    def audit_initiator(self):
        """A pyCADF initiator describing the current authenticated context.

        As a property.
        """
        return notifications.build_audit_initiator()

    @staticmethod
    def query_filter_is_true(filter_name):
        """Determine if bool query param is 'True'.

        We treat this the same way as we do for policy
        enforcement:

        {bool_param}=0 is treated as False

        Any other value is considered to be equivalent to
        True, including the absence of a value (but existence
        as a parameter).

        False Examples for param named `p`:

           * http://host/url
           * http://host/url?p=0

        All other forms of the param 'p' would be result in a True value
        including: `http://host/url?param`.
        """
        val = False
        if filter_name in flask.request.args:
            filter_value = flask.request.args.get(filter_name)
            if (isinstance(filter_value, six.string_types) and
                    filter_value == '0'):
                val = False
            else:
                val = True
        return val

    @property
    def request_body_json(self):
        return flask.request.get_json(silent=True, force=True) or {}

    @staticmethod
    def build_driver_hints(supported_filters):
        """Build list hints based on the context query string.

        :param supported_filters: list of filters supported, so ignore any
                                  keys in query_dict that are not in this list.

        """
        hints = driver_hints.Hints()

        if not flask.request.args:
            return hints

        for key, value in flask.request.args.items(multi=True):
            # Check if this is an exact filter
            if supported_filters is None or key in supported_filters:
                hints.add_filter(key, value)
                continue

            # Check if it is an inexact filter
            for valid_key in supported_filters:
                # See if this entry in query_dict matches a known key with an
                # inexact suffix added.  If it doesn't match, then that just
                # means that there is no inexact filter for that key in this
                # query.
                if not key.startswith(valid_key + '__'):
                    continue

                base_key, comparator = key.split('__', 1)

                # We map the query-style inexact of, for example:
                #
                # {'email__contains', 'myISP'}
                #
                # into a list directive add filter call parameters of:
                #
                # name = 'email'
                # value = 'myISP'
                # comparator = 'contains'
                # case_sensitive = True

                case_sensitive = True
                if comparator.startswith('i'):
                    case_sensitive = False
                    comparator = comparator[1:]
                hints.add_filter(base_key, value,
                                 comparator=comparator,
                                 case_sensitive=case_sensitive)

        # NOTE(henry-nash): If we were to support pagination, we would pull any
        # pagination directives out of the query_dict here, and add them into
        # the hints list.
        return hints

    @classmethod
    def limit(cls, refs, hints):
        """Limit a list of entities.

        The underlying driver layer may have already truncated the collection
        for us, but in case it was unable to handle truncation we check here.

        :param refs: the list of members of the collection
        :param hints: hints, containing, among other things, the limit
                      requested

        :returns: boolean indicating whether the list was truncated, as well
                  as the list of (truncated if necessary) entities.

        """
        NOT_LIMITED = False
        LIMITED = True

        if hints is None or hints.limit is None:
            # No truncation was requested
            return NOT_LIMITED, refs

        if hints.limit.get('truncated', False):
            # The driver did truncate the list
            return LIMITED, refs

        if len(refs) > hints.limit['limit']:
            # The driver layer wasn't able to truncate it for us, so we must
            # do it here
            return LIMITED, refs[:hints.limit['limit']]

        return NOT_LIMITED, refs

    @classmethod
    def _normalize_dict(cls, d):
        return {cls._normalize_arg(k): v for (k, v) in d.items()}

    @staticmethod
    def _normalize_arg(arg):
        return arg.replace(':', '_').replace('-', '_')

    @classmethod
    def _get_domain_id_for_list_request(cls):
        """Get the domain_id for a v3 list call.

        If we running with multiple domain drivers, then the caller must
        specify a domain_id either as a filter or as part of the token scope.

        """
        if not CONF.identity.domain_specific_drivers_enabled:
            # We don't need to specify a domain ID in this case
            return

        domain_id = flask.request.args.get('domain_id')
        if domain_id:
            return domain_id

        token_ref = cls.get_token_ref()

        if token_ref.domain_scoped:
            return token_ref.domain_id
        elif token_ref.project_scoped:
            return token_ref.project_domain['id']
        elif token_ref.system_scoped:
            return
        else:
            msg = 'No domain information specified as part of list request'
            tr_msg = _('No domain information specified as part of list '
                       'request')
            LOG.warning(msg)
            raise exception.Unauthorized(tr_msg)

    @classmethod
    def get_token_ref(cls):
        """Retrieve KeystoneToken object from the auth context and returns it.

        :raises keystone.exception.Unauthorized: If auth context cannot be
                                                 found.
        :returns: The KeystoneToken object.
        """
        try:
            # Retrieve the auth context that was prepared by
            # AuthContextMiddleware.

            auth_context = flask.request.environ.get(
                authorization.AUTH_CONTEXT_ENV, {})
            return auth_context['token']
        except KeyError:
            LOG.warning("Couldn't find the auth context.")
            raise exception.Unauthorized()

    @classmethod
    def _normalize_domain_id(cls, ref):
        """Fill in domain_id if not specified in a v3 call."""
        if not ref.get('domain_id'):
            oslo_ctx = flask.request.environ.get(
                context.REQUEST_CONTEXT_ENV, None)
            if oslo_ctx and oslo_ctx.domain_id:
                # Domain Scoped Token Scenario.
                ref['domain_id'] = oslo_ctx.domain_id
            elif oslo_ctx.is_admin:
                # Legacy "shared" admin token Scenario
                raise exception.ValidationError(
                    _('You have tried to create a resource using the admin '
                      'token. As this token is not within a domain you must '
                      'explicitly include a domain for this resource to '
                      'belong to.'))
            else:
                # TODO(henry-nash): We should issue an exception here since if
                # a v3 call does not explicitly specify the domain_id in the
                # entity, it should be using a domain scoped token.  However,
                # the current tempest heat tests issue a v3 call without this.
                # This is raised as bug #1283539.  Once this is fixed, we
                # should remove the line below and replace it with an error.
                #
                # Ahead of actually changing the code to raise an exception, we
                # issue a deprecation warning.
                versionutils.report_deprecated_feature(
                    LOG,
                    'Not specifying a domain during a create user, group or '
                    'project call, and relying on falling back to the '
                    'default domain, is deprecated as of Liberty. There is no '
                    'plan to remove this compatibility, however, future API '
                    'versions may remove this, so please specify the domain '
                    'explicitly or use a domain-scoped token.')
                ref['domain_id'] = CONF.identity.default_domain_id
        return ref


def base_url(path=''):
    url = CONF['public_endpoint']

    if url:
        substitutions = dict(
            itertools.chain(CONF.items(), CONF.eventlet_server.items()))

        url = url % substitutions
    elif flask.request.environ:
        url = wsgiref.util.application_uri(flask.request.environ)
        # remove version from the URL as it may be part of SCRIPT_NAME but
        # it should not be part of base URL
        url = re.sub(r'/v(3|(2\.0))/*$', '', url)

        # now remove the standard port
        url = utils.remove_standard_port(url)
    else:
        # if we don't have enough information to come up with a base URL,
        # then fall back to localhost. This should never happen in
        # production environment.
        url = 'http://localhost:%d' % CONF.eventlet_server.public_port

    if path:
        # Cleanup leading /v3 if needed.
        path = path.rstrip('/').lstrip('/')
        if path.startswith('v3'):
            path = path[2:].lstrip('/')

    url = url.rstrip('/')
    url = '/'.join([p for p in (url, 'v3', path) if p])
    return url


def full_url(path=''):
    subs = {'url': base_url(path), 'query_string': ''}
    qs = flask.request.environ.get('QUERY_STRING')
    if qs:
        subs['query_string'] = '?%s' % qs
    return '%(url)s%(query_string)s' % subs


def set_unenforced_ok():
    # Does the work for unenforced_api. This must be used outside of a
    # decorator in some limited, such as when a ValidationError is raised up
    # from a "before_request" function (body_json checker is a prime example)
    setattr(flask.g, enforcer._ENFORCEMENT_CHECK_ATTR, True)


def unenforced_api(f):
    """Decorate a resource method to mark is as an unenforced API.

    Explicitly exempts an API from receiving the enforced API check,
    specifically for cases such as user self-service password changes (or
    other APIs that must work without already having a token).

    This decorator may also be used if the API has extended enforcement
    logic/varying enforcement logic (such as some of the AUTH paths) where
    the full enforcement will be implemented directly within the methods.
    """
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        set_unenforced_ok()
        return f(*args, **kwargs)
    return wrapper



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sqlalchemy

from keystone.common import sql
from keystone.tests import unit
from keystone.tests.unit import default_fixtures
from keystone.tests.unit.ksfixtures import database


class BaseBackendSqlTests(unit.SQLDriverOverrides, unit.TestCase):

    def setUp(self):
        super(BaseBackendSqlTests, self).setUp()
        self.useFixture(database.Database())
        self.load_backends()

        # populate the engine with tables & fixtures
        self.load_fixtures(default_fixtures)
        # defaulted by the data load
        self.user_foo['enabled'] = True

    def config_files(self):
        config_files = super(BaseBackendSqlTests, self).config_files()
        config_files.append(unit.dirs.tests_conf('backend_sql.conf'))
        return config_files


class BaseBackendSqlModels(BaseBackendSqlTests):

    def load_table(self, name):
        table = sqlalchemy.Table(name,
                                 sql.ModelBase.metadata,
                                 autoload=True)
        return table

    def assertExpectedSchema(self, table, cols):
        table = self.load_table(table)
        for col, type_, length in cols:
            self.assertIsInstance(table.c[col].type, type_)
            if length:
                self.assertEqual(length, table.c[col].type.length)



# Copyright 2018 SUSE Linux GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc

import six

from keystone import exception


@six.add_metaclass(abc.ABCMeta)
class ApplicationCredentialDriverBase(object):

    @abc.abstractmethod
    def authenticate(self, application_credential_id, secret):
        """Validate an application credential.

        :param str application_credential_id: Application Credential ID
        :param str secret: Secret

        :raises AssertionError: If id or secret is invalid.
        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def create_application_credential(self, application_credential, roles):
        """Create a new application credential.

        :param dict application_credential: Application Credential data
        :param list roles: A list of roles that apply to the
                           application_credential.
        :returns: a new application credential
        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def get_application_credential(self, application_credential_id):
        """Get an application credential by the credential id.

        :param str application_credential_id: Application Credential ID
        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def list_application_credentials_for_user(self, user_id, hints):
        """List application credentials for a user.

        :param str user_id: User ID
        :param hints: contains the list of filters yet to be satisfied.
                      Any filters satisfied here will be removed so that
                      the caller will know if any filters remain.
        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def delete_application_credential(self, application_credential_id):
        """Delete a single application credential.

        :param str application_credential_id: ID of the application credential
            to delete.

        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def delete_application_credentials_for_user(self, user_id):
        """Delete all application credentials for a user.

        :param user_id: ID of a user to whose application credentials should
            be deleted.

        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def delete_application_credentials_for_user_on_project(self, user_id,
                                                           project_id):
        """Delete all application credentials for a user on a given project.

        :param str user_id: ID of a user to whose application credentials
            should be deleted.
        :param str project_id: ID of a project on which to filter application
            credentials.

        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def get_access_rule(self, access_rule_id):
        """Get an access rule by its ID.

        :param str access_rule_id: Access Rule ID
        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def list_access_rules_for_user(self, user_id):
        """List the access rules that a user has created.

        Access rules are only created as attributes of application credentials,
        they cannot be created independently.

        :param str user_id: User ID
        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def delete_access_rule(self, access_rule_id):
        """Delete one access rule.

        :param str access_rule_id: Access Rule ID
        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def delete_access_rules_for_user(self, user_id):
        """Delete all access rules for user.

        This is called when the user itself is deleted.

        :param str user_id: User ID
        """
        raise exception.NotImplemented()  # pragma: no cover



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import versionutils

from keystone.conf import utils


dead_retry = cfg.IntOpt(
    'dead_retry',
    default=5 * 60,
    help=utils.fmt("""
Number of seconds memcached server is considered dead before it is tried again.
This is used by the key value store system.
"""))

socket_timeout = cfg.IntOpt(
    'socket_timeout',
    default=3,
    deprecated_for_removal=True,
    deprecated_reason='This option is duplicated with oslo.cache. '
                      'Configure ``keystone.conf [cache] memcache_socket_timeout`` '
                      'option to set the socket_timeout of memcached instead. ',
    deprecated_since=versionutils.deprecated.TRAIN,
    help=utils.fmt("""
Timeout in seconds for every call to a server. This is used by the key value
store system.
"""))

pool_maxsize = cfg.IntOpt(
    'pool_maxsize',
    default=10,
    help=utils.fmt("""
Max total number of open connections to every memcached server. This is used by
the key value store system.
"""))

pool_unused_timeout = cfg.IntOpt(
    'pool_unused_timeout',
    default=60,
    help=utils.fmt("""
Number of seconds a connection to memcached is held unused in the pool before
it is closed. This is used by the key value store system.
"""))

pool_connection_get_timeout = cfg.IntOpt(
    'pool_connection_get_timeout',
    default=10,
    help=utils.fmt("""
Number of seconds that an operation will wait to get a memcache client
connection. This is used by the key value store system.
"""))


GROUP_NAME = __name__.split('.')[-1]
ALL_OPTS = [
    dead_retry,
    socket_timeout,
    pool_maxsize,
    pool_unused_timeout,
    pool_connection_get_timeout,
]


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=GROUP_NAME)


def list_opts():
    return {GROUP_NAME: ALL_OPTS}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import migrate
import sqlalchemy as sql
from sqlalchemy import func


def upgrade(migrate_engine):
    meta = sql.MetaData()
    meta.bind = migrate_engine

    user_table = sql.Table('user', meta, autoload=True)
    local_user_table = sql.Table('local_user', meta, autoload=True)
    password_table = sql.Table('password', meta, autoload=True)

    # migrate data to local_user table
    local_user_values = []
    for row in user_table.select().execute():
        # skip the row that already exists in `local_user`, this could
        # happen if run into a partially-migrated table due to the
        # bug #1549705.
        filter_by = local_user_table.c.user_id == row['id']
        user_count = sql.select([func.count()]).select_from(
            local_user_table).where(filter_by).execute().fetchone()[0]
        if user_count == 0:
            local_user_values.append({'user_id': row['id'],
                                      'domain_id': row['domain_id'],
                                      'name': row['name']})
    if local_user_values:
        local_user_table.insert().values(local_user_values).execute()

    # migrate data to password table
    sel = (
        sql.select([user_table, local_user_table], use_labels=True)
           .select_from(user_table.join(local_user_table, user_table.c.id ==
                                        local_user_table.c.user_id))
    )
    user_rows = sel.execute()
    password_values = []
    for row in user_rows:
        if row['user_password']:
            password_values.append({'local_user_id': row['local_user_id'],
                                    'password': row['user_password']})
    if password_values:
        password_table.insert().values(password_values).execute()

    # NOTE(gnuoy): the `domain_id` unique constraint is not guaranteed to
    # be a fixed name, such as 'ixu_user_name_domain_id`, so we need to
    # search for the correct constraint that only affects
    # user_table.c.domain_id and drop that constraint. (Fix based on
    # morganfainbergs fix in 088_domain_specific_roles.py)
    to_drop = None
    if migrate_engine.name == 'mysql':
        for index in user_table.indexes:
            if (index.unique and len(index.columns) == 2 and
                    'domain_id' in index.columns and 'name' in index.columns):
                to_drop = index
                break
    else:
        for index in user_table.constraints:
            if (len(index.columns) == 2 and 'domain_id' in index.columns and
                    'name' in index.columns):
                to_drop = index
                break
    # remove domain_id and name unique constraint
    if migrate_engine.name != 'sqlite' and to_drop is not None:
        migrate.UniqueConstraint(user_table.c.domain_id,
                                 user_table.c.name,
                                 name=to_drop.name).drop()

    # drop user columns
    user_table.c.domain_id.drop()
    user_table.c.name.drop()
    user_table.c.password.drop()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import os

import fixtures
from oslo_db import options as db_options
from oslo_db.sqlalchemy import enginefacade

from keystone.common import sql
import keystone.conf
from keystone.tests import unit


CONF = keystone.conf.CONF


def run_once(f):
    """A decorator to ensure the decorated function is only executed once.

    The decorated function is assumed to have a one parameter.

    """
    @functools.wraps(f)
    def wrapper():
        if not wrapper.already_ran:
            f()
            wrapper.already_ran = True
    wrapper.already_ran = False
    return wrapper


# NOTE(I159): Every execution all the options will be cleared. The method must
# be called at the every fixture initialization.
def initialize_sql_session(connection_str=unit.IN_MEM_DB_CONN_STRING,
                           enforce_sqlite_fks=True):
    # Make sure the DB is located in the correct location, in this case set
    # the default value, as this should be able to be overridden in some
    # test cases.
    db_options.set_defaults(
        CONF,
        connection=connection_str)

    # Enable the  Sqlite FKs for global engine by default.
    facade = enginefacade.writer
    engine = facade.get_engine()
    f_key = 'ON' if enforce_sqlite_fks else 'OFF'
    if engine.name == 'sqlite':
        engine.connect().execute('PRAGMA foreign_keys = ' + f_key)


@run_once
def _load_sqlalchemy_models():
    """Find all modules containing SQLAlchemy models and import them.

    This creates more consistent, deterministic test runs because tables
    for all core and extension models are always created in the test
    database. We ensure this by importing all modules that contain model
    definitions.

    The database schema during test runs is created using reflection.
    Reflection is simply SQLAlchemy taking the model definitions for
    all models currently imported and making tables for each of them.
    The database schema created during test runs may vary between tests
    as more models are imported. Importing all models at the start of
    the test run avoids this problem.

    """
    keystone_root = os.path.normpath(os.path.join(
        os.path.dirname(__file__), '..', '..', '..'))
    for root, dirs, files in os.walk(keystone_root):
        # NOTE(morganfainberg): Slice the keystone_root off the root to ensure
        # we do not end up with a module name like:
        # Users.home.openstack.keystone.assignment.backends.sql
        root = root[len(keystone_root):]
        if root.endswith('backends') and 'sql.py' in files:
            # The root will be prefixed with an instance of os.sep, which will
            # make the root after replacement '.<root>', the 'keystone' part
            # of the module path is always added to the front
            module_root = ('keystone.%s' %
                           root.replace(os.sep, '.').lstrip('.'))
            module_components = module_root.split('.')
            module_without_backends = ''
            for x in range(0, len(module_components) - 1):
                module_without_backends += module_components[x] + '.'
            module_without_backends = module_without_backends.rstrip('.')

            module_name = module_root + '.sql'
            __import__(module_name)


class Database(fixtures.Fixture):
    """A fixture for setting up and tearing down a database."""

    def __init__(self):
        super(Database, self).__init__()
        initialize_sql_session()
        _load_sqlalchemy_models()
        sql.enable_sqlite_foreign_key()

    def setUp(self):
        super(Database, self).setUp()

        with sql.session_for_write() as session:
            self.engine = session.get_bind()
        self.addCleanup(sql.cleanup)
        sql.ModelBase.metadata.create_all(bind=self.engine)
        self.addCleanup(sql.ModelBase.metadata.drop_all, bind=self.engine)

    def recreate(self):
        sql.ModelBase.metadata.create_all(bind=self.engine)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from keystone.conf import utils


debug_middlware = cfg.BoolOpt(
    'debug_middleware',
    default=False,
    help=utils.fmt("""
If set to true, this enables the oslo debug middleware in Keystone. This
Middleware prints a lot of information about the request and the response. It
is useful for getting information about the data on the wire (decoded) and
passed to the WSGI application pipeline. This middleware has no effect on
the "debug" setting in the [DEFAULT] section of the config file or setting
Keystone's log-level to "DEBUG"; it is specific to debugging the WSGI data
as it enters and leaves Keystone (specific request-related data). This option
is used for introspection on the request and response data between the web
server (apache, nginx, etc) and Keystone.

This middleware is inserted as the first element in the middleware chain
and will show the data closest to the wire.

WARNING: NOT INTENDED FOR USE IN PRODUCTION. THIS MIDDLEWARE CAN AND WILL EMIT
SENSITIVE/PRIVILEGED DATA.
"""))

GROUP_NAME = __name__.split('.')[-1]
ALL_OPTS = [
    debug_middlware,
]


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=GROUP_NAME)


def list_opts():
    return {GROUP_NAME: ALL_OPTS}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import versionutils

from keystone.conf import utils


_DEPRECATED_MSG = utils.fmt("""
This option has been superseded by ephemeral users existing in the domain
of their identity provider.
""")

driver = cfg.StrOpt(
    'driver',
    default='sql',
    help=utils.fmt("""
Entry point for the federation backend driver in the `keystone.federation`
namespace. Keystone only provides a `sql` driver, so there is no reason to set
this option unless you are providing a custom entry point.
"""))

assertion_prefix = cfg.StrOpt(
    'assertion_prefix',
    default='',
    help=utils.fmt("""
Prefix to use when filtering environment variable names for federated
assertions. Matched variables are passed into the federated mapping engine.
"""))

remote_id_attribute = cfg.StrOpt(
    'remote_id_attribute',
    help=utils.fmt("""
Default value for all protocols to be used to obtain the entity ID of the
Identity Provider from the environment. For `mod_shib`, this would be
`Shib-Identity-Provider`. For `mod_auth_openidc`, this could be
`HTTP_OIDC_ISS`. For `mod_auth_mellon`, this could be `MELLON_IDP`. This can be
overridden on a per-protocol basis by providing a `remote_id_attribute` to the
federation protocol using the API.
"""))

federated_domain_name = cfg.StrOpt(
    'federated_domain_name',
    default='Federated',
    deprecated_for_removal=True,
    deprecated_reason=_DEPRECATED_MSG,
    deprecated_since=versionutils.deprecated.TRAIN,
    help=utils.fmt("""
An arbitrary domain name that is reserved to allow federated ephemeral users to
have a domain concept. Note that an admin will not be able to create a domain
with this name or update an existing domain to this name. You are not advised
to change this value unless you really have to.
"""))

trusted_dashboard = cfg.MultiStrOpt(
    'trusted_dashboard',
    default=[],
    help=utils.fmt("""
A list of trusted dashboard hosts. Before accepting a Single Sign-On request to
return a token, the origin host must be a member of this list. This
configuration option may be repeated for multiple values. You must set this in
order to use web-based SSO flows. For example:
trusted_dashboard=https://acme.example.com/auth/websso
trusted_dashboard=https://beta.example.com/auth/websso
"""))

sso_callback_template = cfg.StrOpt(
    'sso_callback_template',
    default='/etc/keystone/sso_callback_template.html',
    help=utils.fmt("""
Absolute path to an HTML file used as a Single Sign-On callback handler. This
page is expected to redirect the user from keystone back to a trusted dashboard
host, by form encoding a token in a POST request. Keystone's default value
should be sufficient for most deployments.
"""))


caching = cfg.BoolOpt(
    'caching',
    default=True,
    help=utils.fmt("""
Toggle for federation caching. This has no effect unless global caching is
enabled. There is typically no reason to disable this.
"""))


GROUP_NAME = __name__.split('.')[-1]
ALL_OPTS = [
    driver,
    assertion_prefix,
    remote_id_attribute,
    federated_domain_name,
    trusted_dashboard,
    sso_callback_template,
    caching,
]


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=GROUP_NAME)


def list_opts():
    return {GROUP_NAME: ALL_OPTS}
